"""
Record field definitions for testing.

This module contains record field definitioons that can be used in 
test context values.
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2018, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

from collections                import OrderedDict

import logging
log = logging.getLogger(__name__)

from annalist.identifiers       import RDF, RDFS, ANNAL
from annalist                   import layout

from annalist.views.fields.render_placement import get_placement_classes
from annalist.views.form_utils.fieldchoice  import FieldChoice

#   -----------------------------------------------------------------------------
#
#   ----- Field definition data
#
#   -----------------------------------------------------------------------------

no_options = [ FieldChoice('', label="(no options)") ]

# ===== Generic/default entity data =====

f_Entity_id = (
    { 'field_id':               "Entity_id"
    , 'field_name':             "entity_id"
    , 'field_label':            "Id"
    , 'field_render_type':      "EntityId"
    , 'field_value_mode':       "Value_direct"
    , 'field_value_type':       "annal:EntityRef"
    , 'field_property_uri':     "annal:id"
    , 'field_placement':        "small:0,12;medium:0,6"
    , 'field_ref_type':         None
    , 'field_ref_field':        None
    , 'field_placeholder':      "(entity id)"
    , 'field_default_value':    ""
    , 'field_choices':          no_options
    })

f_Entity_ref = (
    { 'field_id':               "Entity_ref"
    , 'field_name':             "entity_ref"
    , 'field_label':            "Entity ref"
    , 'field_render_type':      "EntityId"
    , 'field_value_mode':       "Value_direct"
    , 'field_value_type':       "annal:EntityRef"
    , 'field_property_uri':     "rdfs:label"
    , 'field_placement':        "small:0,12;medium:0,6"
    , 'field_ref_type':         None
    , 'field_ref_field':        None
    , 'field_placeholder':      "(entity ref)"
    , 'field_default_value':    ""
    , 'field_choices':          no_options
    })

f_Entity_type = (
    { 'field_id':               "Entity_type"
    , 'field_name':             "entity_type"
    , 'field_label':            "Type"
    , 'field_render_type':      "EntityTypeId"
    , 'field_value_mode':       "Value_direct"
    , 'field_value_type':       "annal:EntityRef"
    , 'field_property_uri':     "annal:type_id"
    , 'field_placement':        "small:0,12;medium:6,6"
    , 'field_ref_type':         "_type"
    , 'field_ref_field':        None
    , 'field_placeholder':      "(type id)"
    , 'field_default_value':    ""
    , 'field_choices':          no_options
    })

f_Entity_label = (
    { 'field_id':               "Entity_label"
    , 'field_name':             "Entity_label"
    , 'field_label':            "Label"
    , 'field_render_type':      "Text"
    , 'field_value_mode':       "Value_direct"
    , 'field_value_type':       "annal:Text"
    , 'field_property_uri':     "rdfs:label"
    , 'field_placement':        "small:0,12"
    , 'field_ref_type':         None
    , 'field_ref_field':        None
    , 'field_placeholder':      "(label)"
    , 'field_default_value':    ""
    , 'field_choices':          no_options
    })

f_Entity_comment = (
    { 'field_id':               "Entity_comment"
    , 'field_name':             "Entity_comment"
    , 'field_label':            "Comment"
    , 'field_render_type':      "Markdown"
    , 'field_value_mode':       "Value_direct"
    , 'field_value_type':       "annal:Richtext"
    , 'field_property_uri':     "rdfs:comment"
    , 'field_placement':        "small:0,12"
    , 'field_ref_type':         None
    , 'field_ref_field':        None
    , 'field_placeholder':      "(description)"
    , 'field_default_value':    ""
    , 'field_choices':          None # no_options
    })

f_Entity_uri = (
    { 'field_id':               "Entity_uri"
    , 'field_name':             "Entity_uri"
    , 'field_label':            "Entity URI"
    , 'field_render_type':      "Identifer"
    , 'field_value_mode':       "Value_direct"
    , 'field_value_type':       "annal:Identifer"
    , 'field_property_uri':     "annal:uri"
    , 'field_placement':        "small:0,12"
    , 'field_ref_type':         None
    , 'field_ref_field':        None
    , 'field_placeholder':      "(URI)"
    , 'field_default_value':    ""
    , 'field_choices':          no_options
    })

f_Entity_see_also_r = (
    { 'field_id':               "Entity_see_also_r"
    , 'field_name':             "Entity_see_also_r"
    , 'field_label':            "See also"
    , 'field_render_type':      "Group_Set_Row"
    , 'field_value_mode':       "Value_direct"
    , 'field_value_type':       "annal:Entity_see_also_list"
    , 'field_property_uri':     "rdfs:seeAlso"
    , 'field_placement':        "small:0,12"
    , 'field_ref_type':         None
    , 'field_ref_field':        None
    , 'field_placeholder':      "(Links to further information)"
    , 'field_default_value':    None
    , 'field_choices':          no_options
    })

f_Entity_see_also = (
    { 'field_id':               "Entity_see_also"
    , 'field_name':             "Entity_see_also"
    , 'field_label':            "Link to further information"
    , 'field_render_type':      "URILink"
    , 'field_value_mode':       "Value_direct"
    , 'field_entity_type':      "annal:Entity_see_also_list"
    , 'field_value_type':       "rdfs:Resource"
    , 'field_property_uri':     "rdfs:seeAlso"
    , 'field_placement':        "small:0,12"
    , 'field_ref_type':         None
    , 'field_ref_field':        None
    , 'field_placeholder':      "(URL for further information)"
    , 'field_default_value':    ""
    , 'field_choices':          no_options
    })

# ===== BibEntry data (used only for testing) =====

f_Bib_type = (
    { 'field_id':               "Bib_type"
    , 'field_name':             "Bib_type"
    , 'field_label':            "Bib_type"
    , 'field_render_type':      "Text"
    , 'field_value_mode':       "Value_direct"
    , 'field_entity_type':      "annal:Type"
    , 'field_value_type':       "annal:Text"
    , 'field_property_uri':     "Bib:type"
    , 'field_placement':        "small:0,12;medium:6,6"
    , 'field_ref_type':         None
    , 'field_ref_field':        None
    , 'field_placeholder':      "(Bib_type)"
    , 'field_default_value':    ""
    , 'field_choices':          no_options
    })

f_Bib_title = (
    { 'field_id':               "Bib_title"
    , 'field_name':             "Bib_title"
    , 'field_label':            "Bib_title"
    , 'field_render_type':      "Text"
    , 'field_value_mode':       "Value_direct"
    , 'field_entity_type':      "annal:Type"
    , 'field_value_type':       "annal:Text"
    , 'field_property_uri':     "Bib:title"
    , 'field_placement':        "small:0,12"
    , 'field_ref_type':         None
    , 'field_ref_field':        None
    , 'field_placeholder':      "(Bib_title)"
    , 'field_default_value':    ""
    , 'field_choices':          no_options
    })

f_Bib_month = (
    { 'field_id':               "Bib_month"
    , 'field_name':             "Bib_month"
    , 'field_label':            "Bib_month"
    , 'field_render_type':      "Text"
    , 'field_value_mode':       "Value_direct"
    , 'field_entity_type':      "annal:Type"
    , 'field_value_type':       "annal:Text"
    , 'field_property_uri':     "Bib:month"
    , 'field_placement':        "small:0,12;medium:0,6"
    , 'field_ref_type':         None
    , 'field_ref_field':        None
    , 'field_placeholder':      "(Bib_month)"
    , 'field_default_value':    ""
    , 'field_choices':          no_options
    })

f_Bib_year = (
    { 'field_id':               "Bib_year"
    , 'field_name':             "Bib_year"
    , 'field_label':            "Bib_year"
    , 'field_render_type':      "Text"
    , 'field_value_mode':       "Value_direct"
    , 'field_entity_type':      "annal:Type"
    , 'field_value_type':       "annal:Text"
    , 'field_property_uri':     "Bib:year"
    , 'field_placement':        "small:0,12;medium:6,6"
    , 'field_ref_type':         None
    , 'field_ref_field':        None
    , 'field_placeholder':      "(Bib_year)"
    , 'field_default_value':    ""
    , 'field_choices':          no_options
    })

f_Bib_authors = (
    { 'field_id':               "Bib_authors"
    , 'field_name':             "Bib_authors"
    , 'field_label':            "Bib_authors"
    , 'field_render_type':      "Text"
    , 'field_value_mode':       "Value_direct"
    , 'field_entity_type':      "annal:Type"
    , 'field_value_type':       "annal:Text"
    , 'field_property_uri':     "Bib:authors"
    , 'field_placement':        "small:0,12"
    , 'field_ref_type':         None
    , 'field_ref_field':        None
    , 'field_placeholder':      "(Bib_authors)"
    , 'field_default_value':    ""
    , 'field_choices':          no_options
    })

f_Bib_editors = (
    { 'field_id':               "Bib_editors"
    , 'field_name':             "Bib_editors"
    , 'field_label':            "Bib_editors"
    , 'field_render_type':      "Text"
    , 'field_value_mode':       "Value_direct"
    , 'field_entity_type':      "annal:Type"
    , 'field_value_type':       "annal:Text"
    , 'field_property_uri':     "Bib:editors"
    , 'field_placement':        "small:0,12"
    , 'field_ref_type':         None
    , 'field_ref_field':        None
    , 'field_placeholder':      "(Bib_editors)"
    , 'field_default_value':    ""
    , 'field_choices':          no_options
    })

f_Bib_journal = (
    { 'field_id':               "Bib_journal"
    , 'field_name':             "Bib_journal"
    , 'field_label':            "Bib_journal"
    , 'field_render_type':      "Text"
    , 'field_value_mode':       "Value_direct"
    , 'field_entity_type':      "annal:Type"
    , 'field_value_type':       "annal:Text"
    , 'field_property_uri':     "Bib:journal"
    , 'field_placement':        "small:0,12"
    , 'field_ref_type':         None
    , 'field_ref_field':        None
    , 'field_placeholder':      "(Bib_journal)"
    , 'field_default_value':    ""
    , 'field_choices':          no_options
    })

f_Bib_bookentry = (
    { 'field_id':               "Bib_bookentry"
    , 'field_name':             "Bib_bookentry"
    , 'field_label':            "Bib_bookentry"
    , 'field_render_type':      "Text"
    , 'field_value_mode':       "Value_direct"
    , 'field_entity_type':      "annal:Type"
    , 'field_value_type':       "annal:Text"
    , 'field_property_uri':     "Bib:bookentry"
    , 'field_placement':        "small:0,12"
    , 'field_ref_type':         None
    , 'field_ref_field':        None
    , 'field_placeholder':      "(Bib_bookentry)"
    , 'field_default_value':    ""
    , 'field_choices':          no_options
    })

f_Bib_publication_details = (
    { 'field_id':               "Bib_publication_details"
    , 'field_name':             "Bib_publication_details"
    , 'field_label':            "Bib_publication_details"
    , 'field_render_type':      "Text"
    , 'field_value_mode':       "Value_direct"
    , 'field_entity_type':      "annal:Type"
    , 'field_value_type':       "annal:Text"
    , 'field_property_uri':     "Bib:publication_details"
    , 'field_placement':        "small:0,12"
    , 'field_ref_type':         None
    , 'field_ref_field':        None
    , 'field_placeholder':      "(Bib_publication_details)"
    , 'field_default_value':    ""
    , 'field_choices':          no_options
    })

f_Bib_identifiers = (
    { 'field_id':               "Bib_identifiers"
    , 'field_name':             "Bib_identifiers"
    , 'field_label':            "Bib_identifiers"
    , 'field_render_type':      "Text"
    , 'field_value_mode':       "Value_direct"
    , 'field_entity_type':      "annal:Type"
    , 'field_value_type':       "annal:Text"
    , 'field_property_uri':     "Bib:identifiers"
    , 'field_placement':        "small:0,12"
    , 'field_ref_type':         None
    , 'field_ref_field':        None
    , 'field_placeholder':      "(Bib_identifiers)"
    , 'field_default_value':    ""
    , 'field_choices':          no_options
    })

f_Bib_license = (
    { 'field_id':               "Bib_license"
    , 'field_name':             "Bib_license"
    , 'field_label':            "Bib_license"
    , 'field_render_type':      "Text"
    , 'field_value_mode':       "Value_direct"
    , 'field_entity_type':      "annal:Type"
    , 'field_value_type':       "annal:Text"
    , 'field_property_uri':     "Bib:license"
    , 'field_placement':        "small:0,12"
    , 'field_ref_type':         None
    , 'field_ref_field':        None
    , 'field_placeholder':      "(Bib_license)"
    , 'field_default_value':    ""
    , 'field_choices':          no_options
    })

f_Bib_note = (
    { 'field_id':               "Bib_note"
    , 'field_name':             "Bib_note"
    , 'field_label':            "Bib_note"
    , 'field_render_type':      "Text"
    , 'field_value_mode':       "Value_direct"
    , 'field_entity_type':      "annal:Type"
    , 'field_value_type':       "annal:Text"
    , 'field_property_uri':     "Bib:note"
    , 'field_placement':        "small:0,12"
    , 'field_ref_type':         None
    , 'field_ref_field':        None
    , 'field_placeholder':      "(Bib_note)"
    , 'field_default_value':    ""
    , 'field_choices':          no_options
    })

# ===== Field data =====

f_Field_id = (
    { "field_id":               "Field_id"
    , "field_name":             "entity_id"
    , "field_label":            "Field Id"
    , "field_render_type":      "EntityId"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Field"
    , "field_value_type":       "annal:EntityRef"
    , "field_property_uri":     "annal:id"
    , "field_placement":        "small:0,12;medium:0,6"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(field id)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_Field_render_type = (
    { "field_id":               "Field_render_type"
    , "field_name":             "Field_render_type"
    , "field_label":            "Render type"
    , "field_render_type":      "Enum_choice"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Field"
    , "field_value_type":       "annal:EntityRef"
    , "field_property_uri":     "annal:field_render_type"
    , "field_placement":        "small:0,12;medium:6,6"
    , "field_ref_type":         "_enum_render_type"
    , "field_ref_field":        None
    , "field_placeholder":      "(field render type)"
    , "field_default_value":    "Text"
    , "field_choices":          []
    })

f_Field_label = (
    { "field_id":               "Field_label"
    , "field_name":             "Field_label"
    , "field_label":            "Field label"
    , "field_render_type":      "Text"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Field"
    , "field_value_type":       "annal:Text"
    , "field_property_uri":     "rdfs:label"
    , "field_placement":        "small:0,12"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(field label)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_Field_help = (
    { "field_id":               "Field_help"
    , "field_name":             "Field_help"
    , "field_label":            "Help"
    , "field_render_type":      "Markdown"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Field"
    , "field_value_type":       "annal:Richtext"
    , "field_property_uri":     "rdfs:comment"
    , "field_placement":        "small:0,12"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(Field usage commentary or help text)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_Field_property = (
    { "field_id":               "Field_property"
    , "field_name":             "Field_property"
    , "field_label":            "Property URI"
    , "field_render_type":      "Identifier"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Field"
    , "field_value_type":       "annal:Identifier"
    , "field_property_uri":     "annal:property_uri"
    , "field_placement":        "small:0,12;medium:0,6"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(field URI or CURIE)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_Field_placement = (
    { "field_id":               "Field_placement"
    , "field_name":             "Field_placement"
    , "field_label":            "Position/size"
    , "field_render_type":      "Placement"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Field"
    , "field_value_type":       "annal:Placement"
    , "field_property_uri":     "annal:field_placement"
    , "field_placement":        "small:0,12;medium:6,6"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(field position and size)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_Field_sel = (
    { "field_id":               "Field_sel"
    , "field_name":             "entity_id"
    , "field_label":            "Field sel"
    , "field_render_type":      "EntityId"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Field"
    , "field_value_type":       "annal:EntityRef"
    , "field_property_uri":     "annal:id"
    , "field_placement":        "small:0,12;medium:0,6"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(field id)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_Field_superproperty_uris = (
    { "field_id":               "Field_superproperty_uris"
    , "field_name":             "Field_superproperty_uris"
    , "field_label":            "Superproperty URIs"
    , "field_render_type":      "Group_Set_Row"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Field"
    , "field_value_type":       "annal:Field_superproperty_uri"
    , "field_property_uri":     "annal:superproperty_uri"
    , "field_placement":        "small:0,12"
    , "field_choices":          no_options
    })

f_Field_value_type = (
    { "field_id":               "Field_value_type"
    , "field_name":             "Field_value_type"
    , "field_label":            "Value type"
    , "field_render_type":      "Identifier"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Field"
    , "field_value_type":       "annal:Identifier"
    , "field_property_uri":     "annal:field_value_type"
    , "field_placement":        "small:0,12;medium:0,6"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(field value type)"
    , "field_default_value":    "annal:Text"
    , "field_choices":          no_options
    })

f_Field_value_mode = (
    { "field_id":               "Field_value_mode"
    , "field_name":             "Field_value_mode"
    , "field_label":            "Value mode"
    , "field_render_type":      "Enum_choice"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Field"
    , "field_value_type":       "annal:EntityRef"
    , "field_property_uri":     "annal:field_value_mode"
    , "field_placement":        "small:0,12;medium:6,6"
    , "field_ref_type":         "_enum_value_mode"
    , "field_ref_field":        None
    , "field_placeholder":      "(field value mode)"
    , "field_default_value":    "Value_direct"
    , "field_choices":          []
    })

f_Field_entity_type = (
    { "field_id":               "Field_entity_type"
    , "field_name":             "Field_entity_type" 
    , "field_label":            "Entity type"
    , "field_render_type":      "Identifier"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Field"
    , "field_value_type":       "annal:Identifier"
    , "field_property_uri":     "annal:field_entity_type"
    , "field_placement":        "small:0,12;medium:0,6"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_default_value":    ""
    , "field_choices":          []
    })

f_Field_typeref = (
    { "field_id":               "Field_typeref"
    , "field_name":             "Field_typeref"
    , "field_label":            "Refer to type"
    , "field_render_type":      "Enum_optional"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Field"
    , "field_value_type":       "annal:EntityRef"
    , "field_property_uri":     "annal:field_ref_type"
    , "field_placement":        "small:0,12;medium:0,6"
    , "field_ref_type":         "_type"
    , "field_ref_field":        None
    , "field_placeholder":      "(no type selected)"
    , "field_default_value":    ""
    , "field_choices":          []
    })

f_Field_fieldref = (
    { "field_id":               "Field_fieldref"
    , "field_name":             "Field_fieldref"
    , "field_label":            "Refer to field"
    , "field_render_type":      "Identifier"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Field"
    , "field_value_type":       "annal:Identifier"
    , "field_property_uri":     "annal:field_ref_field"
    , "field_placement":        "small:0,12;medium:6,6"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(field URI or CURIE)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_Field_default = (
    { "field_id":               "Field_default"
    , "field_name":             "Field_default"
    , "field_label":            "Default value"
    , "field_render_type":      "Text"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Field"
    , "field_value_type":       "annal:Text"
    , "field_property_uri":     "annal:default_value"
    , "field_placement":        "small:0,12"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(field default value)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_Field_placeholder = (
    { "field_id":               "Field_placeholder"
    , "field_name":             "Field_placeholder"
    , "field_label":            "Placeholder"
    , "field_render_type":      "Text"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Field"
    , "field_value_type":       "annal:Text"
    , "field_property_uri":     "annal:placeholder"
    , "field_placement":        "small:0,12"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(placeholder text)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_Field_tooltip = (
    { "field_id":               "Field_tooltip"
    , "field_name":             "Field_tooltip"
    , "field_label":            "Tooltip"
    , "field_render_type":      "Textarea"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Field"
    , "field_value_type":       "annal:Longtext"
    , "field_property_uri":     "annal:tooltip"
    , "field_placement":        "small:0,12"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(Field usage popup help text)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_Field_fields = (
    { "field_id":               "Field_fields"
    , "field_name":             "Field_fields"
    , "field_label":            "Subfields"
    , "field_render_type":      "Group_Seq_Row"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Field"
    , "field_value_type":       "annal:Field_list"
    , "field_property_uri":     "annal:field_fields"
    , "field_placement":        "small:0,12"
    , "field_choices":          no_options
    })

f_Field_repeat_label_add = (
    { "field_id":               "Field_repeat_label_add"
    , "field_name":             "Field_repeat_label_add"
    , "field_label":            "Add value label"
    , "field_render_type":      "Text"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Field"
    , "field_value_type":       "annal:Text"
    , "field_property_uri":     "annal:repeat_label_add"
    , "field_placement":        "small:0,12;medium:0,6"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_default_value":    "Add"
    , "field_choices":          no_options
    })

f_Field_repeat_label_delete = (
    { "field_id":               "Field_repeat_label_delete"
    , "field_name":             "Field_repeat_label_delete"
    , "field_label":            "Remove value label"
    , "field_render_type":      "Text"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Field"
    , "field_value_type":       "annal:Text"
    , "field_property_uri":     "annal:repeat_label_delete"
    , "field_placement":        "small:0,12;medium:6,6"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_default_value":    "Remove"
    , "field_choices":          no_options
    })

f_Field_restrict = (
    { "field_id":               "Field_restrict"
    , "field_name":             "Field_restrict"
    , "field_label":            "Value restriction"
    , "field_render_type":      "Text"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Field"
    , "field_value_type":      "annal:Text"
    , "field_property_uri":     "annal:field_ref_restriction"
    , "field_placement":        "small:0,12"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_Field_undefined = (
    { "field_id":               "Field_undefined"
    , "field_name":             "Field_undefined"
    , "field_label":            "@@undefined@@"
    , "field_render_type":      "@@undefined@@"
    , "field_value_mode":       "@@undefined@@"
    , "field_entity_type":      "annal:Field"
    , "field_value_type":       "@@undefined@@"
    , "field_property_uri":     "@@undefined@@"
    , "field_placement":        "small:0,12"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

# ===== List data =====

f_List_id = (
    { "field_id":               "List_id"
    , "field_name":             "entity_id"
    , "field_label":            "List Id"
    , "field_render_type":      "EntityId"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:List"
    , "field_value_type":       "annal:EntityRef"
    , "field_property_uri":     "annal:id"
    , "field_placement":        "small:0,12;medium:0,6"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(list id)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_List_type = (
    { "field_id":               "List_type"
    , "field_name":             "List_type"
    , "field_label":            "List display type"
    , "field_render_type":      "Enum_choice"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:List"
    , "field_value_type":       "annal:List_type"
    , "field_property_uri":     "annal:display_type"
    , "field_placement":        "small:0,12;medium:6,6"
    , "field_ref_type":         "_enum_list_type"
    , "field_ref_field":        None
    , "field_placeholder":      "(list type)"
    , "field_default_value":    "List"
    , "field_choices":          no_options
    })

f_List_label = (
    { "field_id":               "List_label"
    , "field_name":             "List_label"
    , "field_label":            "Label"
    , "field_render_type":      "Text"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:List"
    , "field_value_type":       "annal:Text"
    , "field_property_uri":     "rdfs:label"
    , "field_placement":        "small:0,12"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(list label)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_List_comment = (
    { "field_id":               "List_comment"
    , "field_name":             "List_comment"
    , "field_value_type":       "annal:Richtext"
    , "field_render_type":      "Markdown"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:List"
    , "field_value_type":       "annal:Richtext"
    , "field_label":            "Help"
    , "field_property_uri":     "rdfs:comment"
    , "field_placement":        "small:0,12"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(description of list view)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_List_default_type = (
    { "field_id":               "List_default_type"
    , "field_name":             "List_default_type"
    , "field_label":            "Default type"
    , "field_render_type":      "Enum_optional"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:List"
    , "field_value_type":       "annal:Type"
    , "field_property_uri":     "annal:default_type"
    , "field_placement":        "small:0,12;medium:0,6"
    , "field_ref_type":         "_type"
    , "field_ref_field":        None
    , "field_placeholder":      "(default entity type)"
    , "field_default_value":    "Default_type"
    , "field_choices":          no_options
    })

f_List_default_view = (
    { "field_id":               "List_default_view"
    , "field_name":             "List_default_view"
    , "field_label":            "Default view"
    , "field_render_type":      "Enum_optional"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:List"
    , "field_value_type":       "annal:View"
    , "field_property_uri":     "annal:default_view"
    , "field_placement":        "small:0,12;medium:6,6"
    , "field_ref_type":         "_view"
    , "field_ref_field":        None
    , "field_placeholder":      "(view id)"
    , "field_default_value":    "Default_view"
    , "field_choices":          no_options
    })

f_List_entity_selector = (
    { "field_id":               "List_entity_selector"
    , "field_name":             "List_entity_selector"
    , "field_label":            "Selector"
    , "field_render_type":      "Text"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:List"
    , "field_value_type":       "annal:Text"
    , "field_property_uri":     "annal:list_entity_selector"
    , "field_placement":        "small:0,12"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(entity selector; e.g. 'ALL', '<type> in [@type]', '[<field>]==<value>', etc.)"
    , "field_default_value":    "ALL"
    , "field_choices":          no_options
    })

f_List_entity_type = (
    { "field_id":               "List_entity_type"
    , "field_name":             "List_entity_type"
    , "field_label":            "List entity type"
    , "field_render_type":      "Identifier"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:List"
    , "field_value_type":       "annal:Identifier"
    , "field_property_uri":     "annal:list_entity_type"
    , "field_placement":        "small:0,12"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(entity type URI/CURIE displayed by list)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_List_fields = (
    { "field_id":               "List_fields"
    , "field_name":             "List_fields"
    , "field_label":            "Fields"
    , "field_render_type":      "Group_Seq_Row"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:List"
    , "field_value_type":       "annal:List_field"
    , "field_property_uri":     "annal:list_fields"
    , "field_placement":        "small:0,12"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(list field description)"
    , "field_default_value":    None
    , "field_choices":          no_options
    })

# ===== Type data =====

f_Type_id = (
    { "field_id":               "Type_id"
    , "field_name":             "entity_id"
    , "field_label":            "Type Id"
    , "field_render_type":      "EntityId"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Type"
    , "field_value_type":       "annal:EntityRef"
    , "field_property_uri":     "annal:id"
    , "field_placement":        "small:0,12;medium:0,6"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(type id)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_Type_label = (
    { "field_id":               "Type_label"
    , "field_name":             "Type_label"
    , "field_label":            "Label"
    , "field_render_type":      "Text"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Type"
    , "field_value_type":       "annal:Text"
    , "field_property_uri":     "rdfs:label"
    , "field_placement":        "small:0,12"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(label)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_Type_comment = (
    { "field_id":               "Type_comment"
    , "field_name":             "Type_comment"
    , "field_label":            "Comment"
    , "field_render_type":      "Markdown"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Type"
    , "field_value_type":       "annal:Richtext"
    , "field_property_uri":     "rdfs:comment"
    , "field_placement":        "small:0,12"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(type description)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_Type_uri = (
    { "field_id":               "Type_uri"
    , "field_name":             "Type_uri"
    , "field_label":            "Type URI"
    , "field_render_type":      "Identifier"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Type"
    , "field_value_type":       "annal:Identifier"
    , "field_property_uri":     "annal:uri"
    , "field_placement":        "small:0,12" #@@@ "small:0,12;medium:0,6"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(Type URI or CURIE)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_Type_supertype_uris = (
    { "field_id":               "Type_supertype_uris"
    , "field_name":             "Type_supertype_uris"
    , "field_label":            "Supertype URIs"
    , "field_render_type":      "Group_Set_Row"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Type"
    , "field_value_type":       "annal:Type_supertype_uri"
    , "field_property_uri":     "annal:supertype_uri"
    , "field_placement":        "small:0,12"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(Supertype URIs or CURIEs)"
    , "field_default_value":    None
    , "field_choices":          no_options
    })

f_Type_view = (
    { "field_id":               "Type_view"
    , "field_name":             "Type_view"
    , "field_label":            "Default view"
    , "field_render_type":      "Enum_optional"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Type"
    , "field_value_type":       "annal:View"
    , "field_property_uri":     "annal:type_view"
    , "field_placement":        "small:0,12;medium:0,6"
    , "field_ref_type":         "_view"
    , "field_ref_field":        None
    , "field_placeholder":      "(view id)"
    , "field_default_value":    "_view/Default_view"
    , "field_choices":          []
    })

f_Type_list = (
    { "field_id":               "Type_list"
    , "field_name":             "Type_list"
    , "field_label":            "Default list"
    , "field_render_type":      "Enum_optional"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Type"
    , "field_value_type":       "annal:List"
    , "field_property_uri":     "annal:type_list"
    , "field_placement":        "small:0,12;medium:6,6"
    , "field_ref_type":         "_list"
    , "field_ref_field":        None
    , "field_placeholder":      "(list id)"
    , "field_default_value":    "_list/Default_list"
    , "field_choices":          []
    })

f_Type_aliases = (
    { "field_id":               "Type_aliases"
    , "field_name":             "Type_aliases"
    , "field_label":            "Field aliases"
    , "field_render_type":      "Group_Seq_Row"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Type"
    , "field_value_type":       "annal:Type_alias"
    , "field_property_uri":     "annal:field_aliases"
    , "field_placement":        "small:0,12"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(field aliases)"
    , "field_default_value":    None
    , "field_choices":          no_options
    })

# ===== View data =====

f_View_id = (
    { "field_id":               "View_id"
    , "field_name":             "entity_id"
    , "field_label":            "View Id"
    , "field_render_type":      "EntityId"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:View"
    , "field_value_type":       "annal:EntityRef"
    , "field_property_uri":     "annal:id"
    , "field_placement":        "small:0,12;medium:0,6"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(view id)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_View_label = (
    { "field_id":               "View_label"
    , "field_name":             "View_label"
    , "field_label":            "Label"
    , "field_render_type":      "Text"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:View"
    , "field_value_type":       "annal:Text"
    , "field_property_uri":     "rdfs:label"
    , "field_placement":        "small:0,12"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(view label)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_View_comment = (
    { "field_id":               "View_comment"
    , "field_name":             "View_comment"
    , "field_label":            "Help"
    , "field_render_type":      "Markdown"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:View"
    , "field_value_type":       "annal:Richtext"
    , "field_property_uri":     "rdfs:comment"
    , "field_placement":        "small:0,12"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(description of record view)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_View_entity_type = (
    { "field_id":               "View_entity_type"
    , "field_name":             "View_entity_type"
    , "field_label":            "View entity type"
    , "field_render_type":      "Identifier"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:View"
    , "field_value_type":       "annal:Identifier"
    , "field_property_uri":     "annal:view_entity_type"
    , "field_placement":        "small:0,12"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(Entity type URI/CURIE displayed by view)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_View_edit_view = (
    { "field_id":               "View_edit_view"
    , "field_name":             "View_edit_view"
    , "field_label":            "Editable view?"
    , "field_render_type":      "CheckBox"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:View"
    , "field_value_type":       "annal:Boolean"
    , "field_property_uri":     "annal:open_view"
    , "field_placement":        "small:0,12;medium:0,6"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(edit view from edit entity form)"
    , "field_default_value":    True
    , "field_choices":          no_options
    })

f_View_fields = (
    { "field_id":               "View_fields"
    , "field_name":             "View_fields"
    , "field_label":            "Fields"
    , "field_render_type":      "Group_Seq_Row"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:View"
    , "field_value_type":       "annal:View_field"
    , "field_property_uri":     "annal:view_fields"
    , "field_placement":        "small:0,12"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(repeat field description)"
    , "field_default_value":    None
    , "field_choices":          no_options
    })

f_View_field_sel = (
    { "field_id":               "View_field_sel"
    , "field_name":             "View_field_sel"
    , "field_label":            "Field ref"
    , "field_render_type":      "Enum_optional"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:View_field"
    , "field_value_type":       "annal:EntityRef"
    , "field_property_uri":     "annal:field_id"
    , "field_placement":        "small:0,12;medium:0,4"
    , "field_ref_type":         "_field"
    , "field_ref_field":        None
    , "field_placeholder":      "(field reference)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_View_field_property = (
    { "field_id":               "View_field_property"
    , "field_name":             "View_field_property"
    , "field_label":            "Property URI"
    , "field_render_type":      "Identifier"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:View_field"
    , "field_value_type":       "annal:Identifier"
    , "field_property_uri":     "annal:property_uri"
    , "field_placement":        "small:0,12;medium:4,4"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(field URI or CURIE)"
    , "field_default_value":    ""
    , "field_choices":          None
    })

f_View_field_placement = (
    { "field_id":               "View_field_placement"
    , "field_name":             "View_field_placement"
    , "field_label":            "Position/size"
    , "field_render_type":      "Placement"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:View_field"
    , "field_value_type":       "annal:Placement"
    , "field_property_uri":     "annal:field_placement"
    , "field_placement":        "small:0,12;medium:8,4"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(field position and size)"
    , "field_default_value":    ""
    , "field_choices":          None
    })

# ===== Namespace vocabulary data =====

f_Vocab_id = (
    { "field_id":               "Vocab_id"
    , "field_name":             "entity_id"
    , "field_label":            "Prefix"
    , "field_render_type":      "EntityId"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Vocabulary"
    , "field_value_type":       "annal:EntityRef"
    , "field_property_uri":     "annal:id"
    , "field_placement":        "small:0,12;medium:0,6"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(vocabulary id)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

f_Vocab_uri = (
    { "field_id":               "Vocab_uri"
    , "field_name":             "Vocab_uri"
    , "field_label":            "Vocabulary URI"
    , "field_render_type":      "URILink"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "annal:Vocabulary"
    , "field_value_type":       "annal:Identifier"
    , "field_property_uri":     "annal:uri"
    , "field_placement":        "small:0,12"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(Vocabulary namespace URI)"
    , "field_default_value":    ""
    , "field_choices":          no_options
    })

# ===== Test case fields =====

f_Test_repeat_field = (
    { "field_id":               "Test_repeat_field"
    , "field_name":             "Test_repeat_field"
    , "field_label":            "Test repeat field label"
    , "field_render_type":      "Group_Seq_Row"
    , "field_value_mode":       "Value_direct"
    # , "field_entity_type":      "test:img_type"
    , "field_value_type":       "annal:Field_group"
    , "field_property_uri":     "test:repeat_fields"
    , "field_placement":        "small:0,12"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(test repeat field)"
    , "group_id":               "Test_repeat_field"
    , "group_label":            "Test repeat field label"
    , "group_add_label":        "Add Test_repeat_field"
    , "group_delete_label":     "Remove Test_repeat_field"
    , "group_field_list":       [ { "annal:field_id":   "Entity_comment" } ]
    })

f_Test_refimg_field = (
    { "field_id":               "Test_refimg_field"
    , "field_name":             "Test_refimg_field"
    , "field_label":            "Image reference"
    , "field_render_type":      "RefMultifield"
    , "field_value_mode":       "Value_entity"
    , "field_entity_type":      "test:img_type"
    , "field_value_type":       "annal:Field_group"
    , "field_property_uri":     "test:ref_image"
    , "field_placement":        "small:0,12;medium:0,6"
    , "field_ref_type":         "img_type"
    , "field_ref_field":        None
    , "field_placeholder":      "(ref image field)"
    , "group_id":               "Test_refimg_field"
    , "group_label":            "Image reference"
    , "group_add_label":        "Add Test_refimg_field"
    , "group_delete_label":     "Remove Test_refimg_field"
    , "group_field_list":       [ { "annal:field_id":   "Test_comment" }
                                , { "annal:field_id":   "Test_image" }
                                ]
    })

f_Test_rptref_field = (
    { "field_id":               "Test_rptref_field"
    , "field_name":             "Test_rptref_field"
    , "field_label":            "Repeat image reference"
    , "field_render_type":      "Group_Seq_Row"
    , "field_value_mode":       "Value_direct"
    # , "field_entity_type":      "test:img_type"
    , "field_value_type":       "annal:Field_group"
    , "field_property_uri":     "test:rpt_image"
    , "field_placement":        "small:0,12"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(repeat image field)"
    , "group_id":               "Test_rptref_field"
    , "group_label":            "Repeat image reference"
    , "group_add_label":        "Add Test_rptref_field"
    , "group_delete_label":     "Remove Test_rptref_field"
    , "group_field_list":       [ { "annal:field_id":   "Test_refimg_field" } ]
    })

f_Test_comment = (
    { "field_id":               "Test_comment"
    , "field_name":             "Test_comment"
    , "field_label":            "View comment field"
    , "field_render_type":      "Markdown"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "test:img_type"
    , "field_value_type":       "annal:Richtext"
    , "field_property_uri":     "rdfs:comment"
    , "field_placement":        "small:0,12;medium:0,6"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(view comment field)"
    , "field_default_value":    None
    , "field_choices":          None
    })

f_Test_image = (
    { "field_id":               "Test_image"
    , "field_name":             "Test_image"
    , "field_label":            "View image field"
    , "field_render_type":      "RefImage"
    , "field_value_mode":       "Value_direct"
    , "field_entity_type":      "test:img_type"
    , "field_value_type":       "annal:Image"
    , "field_property_uri":     "test:image"
    , "field_placement":        "small:0,12;medium:0,6"
    , "field_ref_type":         None
    , "field_ref_field":        None
    , "field_placeholder":      "(view image field)"
    , "field_default_value":    None
    , "field_choices":          None
    })

# ===== Lookup table =====

fields_by_id = (
    { '_sentinel_not_used':         None
    , "Entity_id":                  f_Entity_id
    , "Entity_ref":                 f_Entity_ref
    , "Entity_type":                f_Entity_type
    , "Entity_label":               f_Entity_label
    , "Entity_comment":             f_Entity_comment
    , "Entity_uri":                 f_Entity_uri
    , "Entity_see_also_r":          f_Entity_see_also_r
    , "Entity_see_also":            f_Entity_see_also
    , "Bib_type":                   f_Bib_type
    , "Bib_title":                  f_Bib_title
    , "Bib_month":                  f_Bib_month
    , "Bib_year":                   f_Bib_year
    , "Bib_authors":                f_Bib_authors
    , "Bib_editors":                f_Bib_editors
    , "Bib_journal":                f_Bib_journal
    , "Bib_bookentry":              f_Bib_bookentry
    , "Bib_publication_details":    f_Bib_publication_details
    , "Bib_identifiers":            f_Bib_identifiers
    , "Bib_license":                f_Bib_license
    , "Bib_note":                   f_Bib_note
    , "Field_id":                   f_Field_id
    , "Field_render_type":          f_Field_render_type
    , "Field_label":                f_Field_label
    , "Field_help":                 f_Field_help
    , "Field_property":             f_Field_property
    , "Field_placement":            f_Field_placement
    , 'Field_superproperty_uris':   f_Field_superproperty_uris
    , "Field_value_type":           f_Field_value_type
    , "Field_value_mode":           f_Field_value_mode
    , "Field_entity_type":          f_Field_entity_type
    , "Field_typeref":              f_Field_typeref
    , "Field_fieldref":             f_Field_fieldref
    , "Field_default":              f_Field_default
    , "Field_placeholder":          f_Field_placeholder
    , "Field_tooltip":              f_Field_tooltip
    , "Field_fields":               f_Field_fields
    , "Field_repeat_label_add":     f_Field_repeat_label_add
    , "Field_repeat_label_delete":  f_Field_repeat_label_delete
    , "Field_restrict":             f_Field_restrict
    , "List_id":                    f_List_id
    , "List_type":                  f_List_type
    , "List_label":                 f_List_label
    , "List_comment":               f_List_comment
    , "List_default_type":          f_List_default_type
    , "List_default_view":          f_List_default_view
    , "List_entity_selector":       f_List_entity_selector
    , "List_entity_type":           f_List_entity_type
    , "List_fields":                f_List_fields
    , "Type_id":                    f_Type_id
    , "Type_label":                 f_Type_label
    , "Type_comment":               f_Type_comment
    , "Type_uri":                   f_Type_uri
    , "Type_supertype_uris":        f_Type_supertype_uris
    , "Type_view":                  f_Type_view
    , "Type_list":                  f_Type_list
    , "Type_aliases":               f_Type_aliases
    , "View_id":                    f_View_id
    , "View_label":                 f_View_label
    , "View_comment":               f_View_comment
    , "View_entity_type":           f_View_entity_type
    , "View_edit_view":             f_View_edit_view
    , "View_fields":                f_View_fields
    , "View_field_sel":             f_View_field_sel
    , "View_field_property":        f_View_field_property
    , "View_field_placement":       f_View_field_placement
    , "Vocab_id":                   f_Vocab_id
    , "Vocab_uri":                  f_Vocab_uri
    , "Test_repeat_field":          f_Test_repeat_field
    , "Test_refimg_field":          f_Test_refimg_field
    , "Test_rptref_field":          f_Test_rptref_field
    , "Test_comment":               f_Test_comment
    , "Test_image":                 f_Test_image
    })

#   -----------------------------------------------------------------------------
#
#   Assemble and return a field definition value
#
#   -----------------------------------------------------------------------------

def get_field_description(field_id, placement=None, name=None, prop_uri=None):
    """
    Returns a `FieldDescription` equivalent dictionary for a supplied field id.
    """
    field_id = field_id.split("/")[-1]
    if field_id in fields_by_id:
        field_desc = fields_by_id[field_id].copy()
    else:
        field_desc = f_Field_undefined.copy()
        field_desc["field_id"]   = "Undefined:"+field_id
        field_desc["field_name"] = "Undefined:"+field_id
    field_pos  = placement or field_desc["field_placement"]
    field_desc["field_placement"]    = get_placement_classes(field_pos)
    field_desc["field_name"]         = name or field_desc["field_name"]
    field_desc["field_property_uri"] = prop_uri or field_desc["field_property_uri"]
    if "group_field_list" in field_desc:
        # Generate "group_field_descs" entry
        group_field_descs = []
        for f in field_desc["group_field_list"]:
            fd = get_field_description(f[ANNAL.CURIE.field_id])
            group_field_descs.append(fd)
        field_desc["group_field_descs"] = group_field_descs
    return field_desc

def get_bound_field(field_id, 
    field_val=None, options=None, placement=None, name=None, prop_uri=None,
    group_add_label=None, group_delete_label=None,
    ):
    """
    Returns a `bound_field` equivalent for a supplied field id and value.
    """
    field_desc  = get_field_description(field_id, 
        placement=placement, name=name, prop_uri=prop_uri
        )
    if group_add_label is not None:
        field_desc["group_add_label"] = group_add_label
    if group_delete_label is not None:
        field_desc["group_delete_label"] = group_delete_label
    f_options   = field_desc.pop('field_choices', None)  # OrderedDict
    options     = options if options is not None else f_options
    bound_field = (
        { "description": field_desc
        })
    if field_val is not None:
        bound_field["field_value"] = field_val
    if isinstance(options, OrderedDict):
        options = list(options.values())
    if options is not None:
        bound_field["options"] = options
    return bound_field

# End.



"""
This module defines a class that is used to gather information about an entity
list or view display that is needed to process various kinds of HTTP request.

The intent of this module is to collect and isolate various housekeeping functions
into a common module to avoid repetition of logic and reduce code clutter in
the various Annalist view processing handlers.
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import json
import re
import traceback

from distutils.version              import LooseVersion

from django.conf                    import settings
from django.http                    import HttpResponse
from django.http                    import HttpResponseRedirect
from django.core.urlresolvers       import resolve, reverse

from utils.py3porting               import urljoin, urlsplit

import annalist
from annalist                       import message
from annalist                       import layout
from annalist.identifiers           import RDF, RDFS, ANNAL
from annalist.util                  import valid_id, extract_entity_id

from annalist.models.entitytypeinfo import (
    EntityTypeInfo, 
    SITE_PERMISSIONS, CONFIG_PERMISSIONS
    )
from annalist.models.collection     import Collection
from annalist.models.recordtype     import RecordType
from annalist.models.recordlist     import RecordList
from annalist.models.recordview     import RecordView
from annalist.models.annalistuser   import default_user_id, unknown_user_id

from annalist.views.confirm         import ConfirmView, dict_querydict
from annalist.views.uri_builder     import (
    uri_param_dict,
    scope_params,
    uri_params, uri_with_params, 
    continuation_url_chain, continuation_chain_url,
    url_update_type_entity_id
    )
from annalist.views.fields.render_entityid  import EntityIdValueMapper


#   -------------------------------------------------------------------------------------------
#
#   Table of authorization symbols added to the display context according to 
#   permissions help by the requesting user
#
#   -------------------------------------------------------------------------------------------

context_authorization_map = (
    { "auth_create":        ["CREATE"]
    , "auth_delete":        ["DELETE"]
    , "auth_update":        ["UPDATE"]
    , "auth_config":        ["CONFIG"]
    , "auth_create_coll":   ["CREATE_COLLECTION", "ADMIN"]
    , "auth_delete_coll":   ["DELETE_COLLECTION", "ADMIN"]
    })

#   -------------------------------------------------------------------------------------------
#
#   Helper functions
#
#   -------------------------------------------------------------------------------------------

def make_data_ref(request_url, data_ref, resource_type=None):
    """
    Returns a URI reference that can be used as a reference a 
    data resource based on the supplied request URL, 
    data resource reference and optional type.

    Scope-related query parameters from the original request_url are 
    preserved, and others are discarded.
    """
    params = scope_params(uri_param_dict(request_url))
    if resource_type:
        params['type'] = resource_type
    return uri_with_params(data_ref, params)

def apply_substitutions(context, text_in):
    """
    Apply substitutions from the supplied `context` to the text supplied as `text_in`,
    returning the resulting string.  This has been introduced to make it easier to 
    create meaningful links in help pages, with values HOST, SITE, COLL and BASE
    added to the context to make references to the current host, site, collection
    and entity base URI respectively.

    Substitutions are made for the following patterns found in `text_in`

    `$name` followed by a non-alphanumeric, non-':' character
    `$name:`
    `$[curie]`
    `$$`

    In the first two, `name` consists of a sequence of alphabetic, numeric or '_' 
    characters, and the pattern is replaced by the corresponding value from the context.

    In the this pattern, `curie` may contain several additional characters that may occur 
    in a compact URI (CURIE); the pattern is replaced by the corresponding context value.

    The final pattern is an escape sequence for inserting a single '$' into the output 
    which might otherwise be treated as a context value substitution.

    If no corresponding context value is found for a substitution pattern, the pattern is
    copied as-is to the output.

    Any other occurrence of '$' (i.e. not part of any pattern above) is untouched.

    >>> context = { 'aa': '-aa-', 'bb': '-bb-', 'c:c': '-cc-'}
    >>> apply_substitutions(context, "foo bar") == 'foo bar'
    True
    >>> apply_substitutions(context, "foo $aa bar") == 'foo -aa- bar'
    True
    >>> apply_substitutions(context, "foo $bb:bar") == 'foo -bb-bar'
    True
    >>> apply_substitutions(context, "foo $[c:c] bar") == 'foo -cc- bar'
    True
    >>> apply_substitutions(context, "foo $$ bar") == 'foo $ bar'
    True
    >>> apply_substitutions(context, "foo $dd bar") == 'foo $dd bar'
    True
    >>> apply_substitutions(context, "foo $ee bar") == 'foo $ee bar'
    True
    >>> apply_substitutions(context, "foo $[f:f] bar") == 'foo $[f:f] bar'
    True
    >>> apply_substitutions(context, "foo $aa $bb: $[c:c] $[f:f] bar") == 'foo -aa- -bb- -cc- $[f:f] bar'
    True
    """
    def sub_fn(matchobj):
        matched = matchobj.group(1) or matchobj.group(2)
        if matchobj.group(3):
            return "$"
        elif matched in context:
            return context[matched]
        return matchobj.group(0)
    namechars  = "_A-Za-z0-9"
    curiechars = "-@.~+*=:;,/?#!"+namechars
    #                          1----)       2----)     3--)
    sub_re     = re.compile("\$([%s]+):?|\$\[([%s]+)\]|\$(\$)"%(namechars, curiechars))
    text_out   = sub_re.sub(sub_fn, text_in)
    return text_out

#   -------------------------------------------------------------------------------------------
#
#   Display information class
#
#   -------------------------------------------------------------------------------------------

class DisplayInfo(object):
    """
    This class collects and organizes common information needed to process various
    kinds of view requests.

    A number of methods are provided that collect different kinds of information,
    allowing the calling method flexibility over what information is actually 
    gathered.  All methods follow a common pattern loosely modeled on an error
    monad, which uses a Django HttpResponse object to record the first problem
    found in the information gathering chain.  Once an error has been detected, 
    subsequent methods do not update the display information, but simply return
    the error response object.

    The information gathering methods do have some dependencies and must be
    invoked in a sequence that ensures the dependencies are satisfied.

    view                is the view object that is being rendered.  This is an instance
                        of a class derived from `AnnalistGenericView`, which in turn is 
                        derived from `django.views.generic.View`.
    action              is the user action for which the form has ben invoked
                        (e.g. "new", "copy", "edit", etc.)
    request_dict        is a dictionary of request parameters
                        For GET requests, this derives from the URI query parameters; 
                        for POST requests it is derived from the submitted form data.
    default_continue    is a default continuation URI to be used when returning from the 
                        current view without an explciitly specified continuation in
                        the request.
    """

    def __init__(self, view, action, request_dict, default_continue):
        self.view               = view
        self.action             = action
        self.is_saved           = False
        self.request_dict       = request_dict
        self.continuation_url   = request_dict.get('continuation_url', None)
        self.default_continue   = default_continue
        # Collection/Type/Entity ids - to be supplied based on form data in POST
        self.orig_coll_id       = None
        self.orig_type_id       = None
        self.orig_entity_id     = None
        self.orig_typeinfo      = None
        self.curr_coll_id       = None
        self.curr_type_id       = None
        self.curr_entity_id     = None
        self.curr_typeinfo      = None
        self.src_entity_id      = None
        # Type-specific messages
        self.type_messages      = None
        # Default no permissions:
        self.authorizations     = dict([(k, False) for k in context_authorization_map])
        self.reqhost            = None
        self.site               = None
        self.sitedata           = None
        self.coll_id            = None
        self.collection         = None
        self.orig_coll          = None  # Original collection for copy
        self.perm_coll          = None  # Collection used for permissions checking
        self.type_id            = None  # Type Id from request URI, not dependent on form data
        self.list_id            = None
        self.recordlist         = None
        self.view_id            = None
        self.recordview         = None
        self.entitydata         = None
        # Response data
        self.http_response      = None
        self.info_messages      = []
        self.error_messages     = []
        return

    def set_orig_coll_id(self, orig_coll_id=None):
        """
        For GET and POST operations, set up details of the collection from which
        an existing identified entity is accessible.  This is used later to check 
        collection access permissions.
        """
        self.orig_coll_id       = EntityIdValueMapper.decode(orig_coll_id)
        # If inherited from another collection, update origin collection object
        if self.orig_coll_id and (self.orig_coll_id != self.coll_id):
            c = Collection.load(self.site, self.orig_coll_id, altscope="all")
            if c:
                self.orig_coll = c
        return self.http_response

    def set_coll_type_entity_id(self,
        orig_coll_id=None,
        orig_type_id=None, orig_entity_id=None,
        curr_type_id=None, curr_entity_id=None
        ):
        """
        For a form POST operation, sets updated collection, type and entity
        identifiers from the form data.

        The original collection id may be different by virtue of inheritance
        from another collection (via 'orig_coll_id' parameter).

        The current type identifier may be different by virtue of the type being
        renamed in the form data (via 'curr_type_id' parameter).
        """
        # log.debug(
        #     "@@ DisplaytInfo.set_coll_type_entity_id: %s/%s/%s -> %s/%s"%
        #       ( orig_coll_id, orig_type_id, orig_entity_id, 
        #         curr_type_id, curr_entity_id
        #       )
        #     )
        self.set_orig_coll_id(orig_coll_id)
        if self.http_response:
            return self.http_response
        self.orig_type_id       = EntityIdValueMapper.decode(orig_type_id)
        self.orig_entity_id     = EntityIdValueMapper.decode(orig_entity_id)
        self.curr_type_id       = EntityIdValueMapper.decode(curr_type_id)
        self.curr_entity_id     = EntityIdValueMapper.decode(curr_entity_id)
        if self.orig_coll_id and (self.orig_coll_id != self.coll_id):
            self.orig_typeinfo = EntityTypeInfo(self.orig_coll, orig_type_id)
        if self.curr_type_id and (self.curr_type_id != self.type_id):
            self.curr_typeinfo = EntityTypeInfo(self.collection, curr_type_id)
        return self.http_response

    def set_messages(self, messages):
        """
        Save type-specific messages for later reporting
        """
        self.type_messages      = messages
        return self.http_response

    def get_site_info(self, reqhost):
        """
        Get site information: site entity object and a copy of the site description data.
        Also saves a copy of the host name to which the current request was directed.
        """
        if not self.http_response:
            self.reqhost        = reqhost
            self.site           = self.view.site(host=reqhost)
            self.sitedata       = self.view.site_data()
        return self.http_response

    def get_coll_info(self, coll_id):
        """
        Check collection identifier, and get reference to collection object.
        """
        assert (self.site is not None)
        if not self.http_response:
            if not Collection.exists(self.site, coll_id):
                self.http_response = self.view.error(
                    dict(self.view.error404values(),
                        message=message.COLLECTION_NOT_EXISTS%{'id': coll_id}
                        )
                    )
            else:
                self.coll_id    = coll_id
                #@@TODO: try with altscope="site"?
                self.collection = Collection.load(self.site, coll_id, altscope="all")
                self.orig_coll  = self.collection
                self.perm_coll  = self.collection
                ver = self.collection.get(ANNAL.CURIE.software_version, None) or "0.0.0"
                if LooseVersion(ver) > LooseVersion(annalist.__version__):
                    self.http_response = self.view.error(
                        dict(self.view.error500values(),
                            message=message.COLLECTION_NEWER_VERSION%{'id': coll_id, 'ver': ver}
                            )
                        )
                self.add_error_messages(self.collection.get_errors())
        return self.http_response

    def flush_collection_caches(self):
        """
        Called to flush collection caches so that changes made independently of 
        the caches can be used.

        NOTE: this is currently called by the top-level collection customize view.
        This is a bit of  a hack to ensure that it is always possible for the user 
        to force caches to be flushed, e.g. when type informatiuon is updated in a 
        different tab or by another user.
        """
        assert (self.collection is not None)
        self.collection.flush_collection_caches()
        return

    def update_coll_version(self):
        """
        Called when an entity has been updated to also update the data version 
        associated with the collection if it was previously created by an older 
        version of Annalist.
        """
        assert (self.collection is not None)
        if not self.http_response:
            self.collection.update_software_compatibility_version()
        return

    def saved(self, is_saved=None):
        """
        Make note that current entity has been saved, and/or return saved status
        """
        if is_saved is not None:
            self.is_saved = is_saved
        return self.is_saved

    def get_type_info(self, type_id):
        """
        Check type identifier, and get a reference to the corresponding type 
        information object.

        This method may be called to override the type id from the original request
        URI, and the DisplayInfo 'type_id' value is not updated so that the value
        from the original request URI is not lost.

        See also method 'get_request_type_info'.
        """
        # print "@@@@ get_type_info: type_id %s"%(type_id,)
        if not self.http_response:
            assert ((self.site and self.collection) is not None)
            if type_id:
                self.curr_typeinfo = EntityTypeInfo(self.collection, type_id)
                self.orig_typeinfo = self.curr_typeinfo
                if not self.curr_typeinfo.recordtype:
                    # log.warning("DisplayInfo.get_type_data: RecordType %s not found"%type_id)
                    self.http_response = self.view.error(
                        dict(self.view.error404values(),
                            message=message.RECORD_TYPE_NOT_EXISTS%(
                                {'id': type_id, 'coll_id': self.coll_id})
                            )
                        )
        return self.http_response

    def get_request_type_info(self, type_id):
        """
        Save and check type identifier from request URI, and get a reference to 
        the corresponding type information object.
        """
        self.type_id = type_id
        return self.get_type_info(type_id)

    def get_list_info(self, list_id):
        """
        Retrieve list definition to use for display
        """
        if not self.http_response:
            assert ((self.site and self.collection) is not None)
            assert list_id
            # log.debug(
            #     "DisplayInfo.get_list_info: collection.get_alt_entities %r"%
            #     [ c.get_id() for c in  self.collection.get_alt_entities(altscope="all") ]
            #     )
            if not self.check_list_id(list_id):
                log.warning("DisplayInfo.get_list_info: RecordList %s not found"%list_id)
                msg1 = message.RECORD_LIST_NOT_EXISTS%{'id': list_id, 'coll_id': self.coll_id}
                self.add_error_message(msg1)
                list_id = self.get_list_id(self.type_id, None)
                msg2 = message.DISPLAY_ALTERNATIVE_LIST%{'id': list_id, 'coll_id': self.coll_id}
                self.add_error_message(msg2)
                #@@
                # self.http_response = self.view.error(
                #     dict(self.view.error404values(),
                #         message=message.RECORD_LIST_NOT_EXISTS%(
                #             {'id': list_id, 'coll_id': self.coll_id})
                #         )
                #     )
                #@@
            self.list_id    = list_id
            self.recordlist = RecordList.load(self.collection, list_id, altscope="all")
            if "@error" in self.recordlist:
                self.http_response = self.view.error(
                    dict(self.view.error500values(),
                        message=message.RECORD_LIST_LOAD_ERROR%(
                            { 'id':       list_id
                            , 'file':     self.recordlist["@error"]
                            , 'message':  self.recordlist["@message"]
                            })
                        )
                    )
            elif self.type_id is None and self.curr_typeinfo is None:
                self.get_type_info(
                    extract_entity_id(self.recordlist[ANNAL.CURIE.default_type])
                    )
            # log.debug("DisplayInfo.get_list_info: %r"%(self.recordlist.get_values()))
        return self.http_response

    def get_view_info(self, view_id):
        """
        Retrieve view definition to use for display
        """
        if not self.http_response:
            assert ((self.site and self.collection) is not None)
            if not self.check_view_id(view_id):
                log.warning("DisplayInfo.get_view_info: RecordView %s not found"%view_id)
                log.warning("Collection: %r"%(self.collection))
                log.warning("Collection._altparent: %r"%(self.collection._altparent))
                # log.warning("\n".join(traceback.format_stack()))
                msg1 = message.RECORD_VIEW_NOT_EXISTS%{'id': view_id, 'coll_id': self.coll_id}
                self.add_error_message(msg1)
                view_id = self.get_view_id(self.type_id, None)
                msg2 = message.DISPLAY_ALTERNATIVE_VIEW%{'id': view_id, 'coll_id': self.coll_id}
                self.add_error_message(msg2)
                #@@
                # self.http_response = self.view.error(
                #     dict(self.view.error404values(),
                #         message=message.RECORD_VIEW_NOT_EXISTS%(
                #             {'id': view_id, 'coll_id': self.coll_id})
                #         )
                #     )
                #@@
            self.view_id    = view_id
            self.recordview = RecordView.load(self.collection, view_id, altscope="all")
            if "@error" in self.recordview:
                self.http_response = self.view.error(
                    dict(self.view.error500values(),
                        message=message.RECORD_VIEW_LOAD_ERROR%(
                            { 'id':       list_id
                            , 'file':     self.recordview["@error"]
                            , 'message':  self.recordview["@message"]
                            })
                        )
                    )
                # log.debug("DisplayInfo.get_view_info: %r"%(self.recordview.get_values()))
        return self.http_response

    def get_entity_info(self, action, entity_id):
        """
        Set up entity id and info to use for display

        Also handles some special case permissions settings if the entity is a Collection.
        """
        if not self.http_response:
            assert self.curr_typeinfo is not None
            self.src_entity_id  = entity_id
            self.curr_entity_id = entity_id
            if action in ["new", "copy"]:
                self.use_entity_id = self.curr_typeinfo.entityclass.allocate_new_id(
                    self.curr_typeinfo.entityparent, base_id=entity_id
                    )
            else:
                self.use_entity_id  = entity_id
            # Special case permissions when accessing collection metadata:
            # use the collection itself rather than the site data collection to which it belongs.
            if self.type_id == "_coll":
                # log.info("DisplayInfo.get_entity_info: access collection data for %s"%entity_id)
                c = Collection.load(self.site, entity_id, altscope="all")
                if c:
                    self.perm_coll = c
        return self.http_response

    # Support methods for response generation

    def check_authorization(self, action):
        """
        Check authorization.  Return None if all is OK, or HttpResonse object.

        Also, save copy of key authorizations for later rendering.
        """
        if not self.http_response:
            # Save key authorizations for later rendering
            for k in context_authorization_map:
                for p in context_authorization_map[k]:
                    self.authorizations[k] = (
                        self.authorizations[k] or 
                        self.view.authorize(p, self.perm_coll) is None
                        )
            # Check requested action
            action = action or "view"
            if self.curr_typeinfo:
                # log.info(
                #     "@@ check_authorization (curr) action %s, type_id %s, entity_id %s"%
                #     (action, self.curr_typeinfo.type_id, self.curr_entity_id)
                #     )
                permissions_map = (
                    self.curr_typeinfo.get_entity_permissions_map(self.curr_entity_id)
                    )
            else:
                # Use site permissions map (some site operations don't have an entity type?)
                permissions_map = SITE_PERMISSIONS
            # Previously, default permission map was applied in view.form_action_auth if no 
            # type-based map was provided.
            self.http_response = (
                self.http_response or 
                self.view.form_action_auth(
                    action, self.perm_coll, permissions_map,
                    continuation_url=self.get_continuation_here()
                    )
                )
        if ( (not self.http_response) and 
             (self.orig_coll_id) and 
             (self.orig_coll_id != self.perm_coll.get_id()) ):
            # Copying content from different collection: check access
            if self.orig_typeinfo:
                # log.info(
                #     "@@ check_authorization (orig) action %s, type_id %s, entity_id %s"%
                #     (action, self.orig_typeinfo.type_id, self.orig_entity_id)
                #     )
                orig_permissions_map = (
                    self.orig_typeinfo.get_entity_permissions_map(self.orig_entity_id)
                    )
            else:
                orig_permissions_map = SITE_PERMISSIONS
            self.http_response = self.view.form_action_auth("view", 
                self.orig_coll, orig_permissions_map,
                continuation_url=self.get_continuation_here()
                )
        return self.http_response

    #@@TODO: not sure if this will be useful...
    # def reset_info_messages(self):
    #     """
    #     Reset informational messages (for form redisplay)
    #     cf. entityedit.form_re_render
    #     """
    #     self.info_messages      = []
    #     return
    #@@

    def add_info_message(self, message):
        """
        Save message to be displayed on successful completion
        """
        self.info_messages.append(message)
        return self.http_response

    def add_error_message(self, message):
        """
        Save error message to be displayed on completion of the current operation
        """
        self.error_messages.append(message)
        return self.http_response

    def add_error_messages(self, messages):
        """
        Save list of error message to be displayed on completion of the current operation
        """
        # print "@@ add_error_messages: > %r"%(messages,)
        self.error_messages.extend(messages)
        # print "@@ add_error_messages: < %r"%(self.error_messages,)
        return self.http_response

    def redisplay_path_params(self):
        """
        Gathers URL details based on the current request URL that can be used 
        to construct a URL to redisplay the current page.

        Returns a pair of values:

            redisplay_path, redisplay_params

        Where 'redisplay_path' is the URL path for the current request,
        and 'redisplay_params' is a selection of request URL parameters that 
        are used to select data for the current display (see 'scope_params').
        """
        redisplay_path   = self.view.get_request_path()
        redisplay_params = scope_params(uri_param_dict(redisplay_path))
        redisplay_params.update(self.get_continuation_url_dict())
        return (redisplay_path, redisplay_params)

    def redirect_response(self, redirect_path, redirect_params={}, action=None):
        """
        Return an HTTP redirect response, with information or warning messages as
        requested included as parameters.

        redirect_path   the URI base path to redirect to.
        redirect_params an optional dictionary of parameters to be applied to the 
                        redirection URI.
        action          an action that must be authorized if the redirection is to occur,
                        otherwise an error is reported and the current page redisplayed.
                        If None or not supplied, no authorization check is performed.

        Returns an HTTP response value.
        """
        # @TODO: refactor existing redirect response code (here and in list/edit modules)
        #        to use this method.  (Look for other instances of HttpResponseRedirect)
        # print "@@ redirect_response: http_response %r"%(self.http_response,)
        if not self.http_response:
            redirect_msg_params = dict(redirect_params)
            if self.info_messages:
                redirect_msg_params.update(self.view.info_params("\n\n".join(self.info_messages)))
            if self.error_messages:
                redirect_msg_params.update(self.view.error_params("\n\n".join(self.error_messages)))
            # print "@@ redirect_response: redirect_msg_params %r"%(redirect_msg_params,)
            redirect_uri = (
                uri_with_params(
                    redirect_path,
                    redirect_msg_params
                    )
                )
            self.http_response = (
                (action and self.check_authorization(action))
                or
                HttpResponseRedirect(redirect_uri)
                )
        return self.http_response

    def display_error_response(self, err_msg):
        """
        Return an HTTP response that redisplays the current view with an error
        message displayed.

        err_msg         is the error message to be displayed.
        """
        redirect_path, redirect_params = self.redisplay_path_params()
        self.add_error_message(err_msg)
        return self.redirect_response(redirect_path, redirect_params=redirect_params)

    def report_error(self, message):
        """
        Set up error response
        """
        log.error(message)
        if not self.http_response:
            self.http_response = self.view.error(
                dict(self.view.error400values(),
                    message=message
                    )
                )
        return self.http_response

    def confirm_delete_entity_response(self, 
            entity_type_id, entity_id, 
            complete_action_uri,
            form_action_field="entity_delete",
            form_value_field="entity_id",
            response_messages = {}
            ):
        """
        This method generates a response when the user has requested deletion
        of an entity from the current collection.  It includes logic to request 
        confirmation of the requested action before proceeding to actually remove
        the entity.

        The request for confirmation is handled via class "ConfirmView" (confirm.py),
        and actual deletion and continuation is performed via the view specified by
        "complete_action_view", which is typically realized by a subclass of 
        "EntityDeleteConfirmedBaseView" (entitydeletebase.py)

        entity_type_id          is the type id of the entity to be deleted.
        entity_id               is the entity id of the entity to be deleted.
        complete_action_uri     identifies a view to be invoked by an HTTP POST operation
                                to complete the entity deletion.
        form_action_field       names the form field that is used to trigger entity deletion
                                in the completion view.  Defaults to "entity_delete"
        form_value_field        names the form field that is used to provide the identifier of
                                the entity or entities to be deleted.
        response_messages       is a dictionary of messages to be used to indicate 
                                various outcomes:
                                "no_entity" is the entity for deletion is not specified.
                                "cannot_delete" if entity does not exist or cannot be deleted.
                                "confirm_completion" to report co,mpletion of entity deletion.
                                If no message dictionary is provided, or if no value is provided 
                                for a particular outcome, a default message value will be used.
                                Messages may use value interpolation for %(type_id)s and %(id)s.
        """
        def _get_message(msgid):
            return response_messages.get(msgid, default_messages.get(msgid)%message_params)

        default_messages = (
            { "no_entity":          message.NO_ENTITY_FOR_DELETE
            , "cannot_delete":      message.CANNOT_DELETE_ENTITY
            , "type_has_values":    message.TYPE_VALUES_FOR_DELETE
            , "confirm_completion": message.REMOVE_ENTITY_DATA
            })
        entity_coll_id = self.collection.get_id()
        message_params = (
            { "id":         entity_id
            , "type_id":    entity_type_id
            , "coll_id":    entity_coll_id
            })

        if not entity_id:
            self.display_error_response(_get_message("no_entity"))
        elif not self.entity_exists(entity_id, entity_type_id):
            self.display_error_response(_get_message("cannot_delete"))
        elif self.entity_is_type_with_values(entity_id, entity_type_id):
            self.display_error_response(_get_message("type_has_values"))

        if not self.http_response:
            # Get user to confirm action before actually doing it
            # log.info(
            #     "entity_coll_id %s, type_id %s, entity_id %s, confirmed_action_uri %s"%
            #     (entity_coll_id, entity_type_id, entity_id, confirmed_action_uri)
            #     )
            delete_params = (
                { form_action_field:    ["Delete"]
                , form_value_field:     [entity_id]
                , "completion_url":     [self.get_continuation_here()]
                , "search_for":         [self.request_dict.get('search_for',"")]
                })
            curi = self.get_continuation_url()
            if curi:
                delete_params["continuation_url"] = [curi]
            return (
                self.check_authorization("delete")
                or
                ConfirmView.render_form(self.view.request,
                    action_description=     _get_message("confirm_completion"),
                    action_params=          dict_querydict(delete_params),
                    confirmed_action_uri=   complete_action_uri,
                    cancel_action_uri=      self.get_continuation_here(),
                    title=                  self.view.site_data()["title"]
                    )
                )
        return self.http_response

    # Additonal support functions for list views

    def get_type_list_id(self, type_id):
        """
        Return default list_id for listing defined type, or None
        """
        list_id = None
        if type_id:
            if self.curr_typeinfo.recordtype:
                list_id = extract_entity_id(
                    self.curr_typeinfo.recordtype.get(ANNAL.CURIE.type_list, None)
                    )
            else:
                log.warning("DisplayInfo.get_type_list_id no type data for %s"%(type_id))
        return list_id

    def check_list_id(self, list_id):
        """
        Check for existence of list definition: 
        if it exists, return the supplied list_id, else None.
        """
        if list_id and RecordList.exists(self.collection, list_id, altscope="all"):
            return list_id
        return None

    def get_list_id(self, type_id, list_id):
        """
        Return supplied list_id if defined, otherwise find default list_id for
        entity type or collection (unless an error has been detected).
        """
        if not self.http_response:
            list_id = (
                list_id or 
                self.check_list_id(self.get_type_list_id(type_id)) or
                self.check_list_id(self.collection.get_default_list()) or
                ("Default_list" if type_id else "Default_list_all")
                )
            if not list_id:
                log.warning("get_list_id failure: %s, type_id %s"%(list_id, type_id))
        return list_id

    def get_list_view_id(self):
        return extract_entity_id(
            self.recordlist.get(ANNAL.CURIE.default_view, None) or "Default_view"
            )

    def get_list_type_id(self):
        return extract_entity_id(
            self.recordlist.get(ANNAL.CURIE.default_type, None) or "Default_type"
            )

    # Additional support functions for collection view

    def get_default_view_type_entity(self):
        """
        Return default view_id, type_id and entity_id to display for collection,
        or None for any values not defined.
        """
        view_id, type_id, entity_id = self.collection.get_default_view()
        return (self.check_view_id(view_id), type_id, entity_id)

    # Additonal support functions for entity views

    def check_view_id(self, view_id):
        """
        Check for existence of view definition: 
        if it exists, return the supplied view_id, else None.
        """
        if view_id and RecordView.exists(self.collection, view_id, altscope="all"):
            return view_id
        return None

    def get_view_id(self, type_id, view_id):
        """
        Get view id or suitable default using type if defined.
        """
        if not self.http_response:
            view_id = (
                view_id or 
                self.check_view_id(self.curr_typeinfo.get_default_view_id())
                )
            if not view_id:
                log.warning("get_view_id: %s, type_id %s"%(view_id, self.type_id))
        return view_id

    def entity_exists(self, entity_id, entity_type):
        """
        Test a supplied entity is defined in the current collection,
        returning true or False.

        entity_id           entity id that is to be tested..
        entity_type         type of entity to test.
        """
        typeinfo = self.curr_typeinfo
        if not typeinfo or typeinfo.get_type_id() != entity_type:
            typeinfo = EntityTypeInfo(self.collection, entity_type)
        return typeinfo.entityclass.exists(typeinfo.entityparent, entity_id)

    def entity_is_type_with_values(self, entity_id, entity_type):
        """
        Test if indicated entity is a type with values defined.
        """
        if entity_type == layout.TYPE_TYPEID:
            typeinfo = EntityTypeInfo(
                self.collection, entity_id
                )
            return next(typeinfo.enum_entity_ids(), None) is not None
        return False

    def get_new_view_uri(self, coll_id, type_id):
        """
        Get URI for entity new view from list display
        """
        return self.view.view_uri(
            "AnnalistEntityNewView", 
            coll_id=coll_id, 
            view_id=self.get_list_view_id(), 
            type_id=type_id,
            action="new"
            )

    def get_edit_view_uri(self, coll_id, type_id, entity_id, action):
        """
        Get URI for entity edit or copy view from list display
        """
        # Use default view for entity type
        # (Use list view id only for new entities)
        return self.view.view_uri(
                "AnnalistEntityDefaultDataView", 
                coll_id=coll_id, 
                type_id=type_id,
                entity_id=entity_id,
                action=action
                )
        # return self.view.view_uri(
        #         "AnnalistEntityEditView", 
        #         coll_id=coll_id, 
        #         view_id=self.get_list_view_id(), 
        #         type_id=type_id,
        #         entity_id=entity_id,
        #         action=action
        #         )

    def get_src_entity_resource_url(self, resource_ref):
        """
        Return URL for accessing source entity resource data 
        (not including any view information contained in the current request URL).

        Contains special logic for accessing collection and site metadata
        """
        assert self.coll_id is not None
        assert self.curr_typeinfo is not None
        type_id   = self.curr_typeinfo.get_type_id()
        if type_id == layout.COLL_TYPEID:
            entity_id = self.src_entity_id or layout.SITEDATA_ID
            base_url  = self.view.get_collection_base_url(entity_id)
        else:
            entity_id = self.src_entity_id or "__unknown_src_entity__"
            base_url  = self.view.get_entity_base_url(
                self.coll_id, type_id,
                entity_id
                )
        return urljoin(base_url, resource_ref)

    # Additonal support functions

    def get_continuation_param(self):
        """
        Return continuation URL specified for the current request, or None.
        """
        cont_here = self.view.continuation_here(self.request_dict, self.default_continue)
        return uri_params({"continuation_url": cont_here})

    def get_continuation_url(self):
        """
        Return continuation URL specified for the current request, or None.
        """
        return self.continuation_url

    def get_continuation_url_dict(self):
        """
        Return dictionary with continuation URL specified for the current request.
        """
        c = self.get_continuation_url()
        return {'continuation_url': c} if c else {}

    def get_continuation_next(self):
        """
        Return continuation URL to be used when returning from the current view.

        Uses the default continuation if no value supplied in request dictionary.
        """
        log.debug(
            "get_continuation_next '%s', default '%s'"%
              (self.continuation_url, self.default_continue)
            )
        return self.continuation_url or self.default_continue

    def get_continuation_here(self, base_here=None):
        """
        Return continuation URL back to the current view.
        """
        # @@TODO: consider merging logic from generic.py, and eliminating method there
        continuation_here = self.view.continuation_here(
            request_dict=self.request_dict,
            default_cont=self.get_continuation_url(),
            base_here=base_here
            )
        # log.info("DisplayInfo.get_continuation_here: %s"%(continuation_here))
        return continuation_here

    def update_continuation_url(self, 
        old_type_id=None, new_type_id=None, 
        old_entity_id=None, new_entity_id=None
        ):
        """
        Update continuation URI to reflect renamed type or entity.
        """
        curi = self.continuation_url
        if curi:
            hops = continuation_url_chain(curi)
            for i in range(len(hops)):
                uribase, params = hops[i]
                uribase = url_update_type_entity_id(uribase, 
                    old_type_id=old_type_id, new_type_id=new_type_id, 
                    old_entity_id=old_entity_id, new_entity_id=new_entity_id
                    )
                hops[i] = (uribase, params)
            curi = continuation_chain_url(hops)
            self.continuation_url = curi
        return curi

    def get_entity_data_ref(self, name_ext=".jsonld", return_type=None):
        """
        Returns a relative reference (from entity base) for the metadata for the
        current entity using the supplied name extension.
        """
        assert self.curr_typeinfo is not None
        data_ref = self.curr_typeinfo.entityclass.meta_resource_name(name_ext=name_ext)
        # log.info("@@@@ get_entity_data_ref: data_ref "+data_ref)
        data_ref = make_data_ref(
            self.view.get_request_path(),   # For parameter values
            data_ref, 
            return_type
            )
        # log.info("@@@@ get_entity_data_ref: data_ref "+data_ref)
        return data_ref

    def get_entity_jsonld_ref(self, return_type=None):
        """
        Returns a relative reference (from entity base) for the metadata for the
        current entity, to be returned as JSON-LD data.
        """
        jsonld_ref = self.get_entity_data_ref(name_ext=".jsonld", return_type=return_type)
        return jsonld_ref

    def get_entity_jsonld_url(self, return_type=None):
        """
        Returns a string that can be used as a reference to the entity metadata resource,
        optionally with a specified type parameter added.

        Extracts appropriate local reference, and combines with entity URL path.
        """
        data_ref = self.get_entity_jsonld_ref(return_type=return_type)
        data_url = self.get_src_entity_resource_url(data_ref)
        log.debug(
            "get_entity_jsonld_url: _entityfile %s, data_ref %s, data_url %s"%
            (self.curr_typeinfo.entityclass._entityfile, data_ref, data_url)
            )
        return data_url

    def get_entity_turtle_ref(self, return_type=None):
        """
        Returns a relative reference (from entity base) for the metadata for the
        current entity, to be returned as Turtle data.
        """
        turtle_ref = self.get_entity_data_ref(name_ext=".ttl", return_type=return_type)
        return turtle_ref

    def get_entity_turtle_url(self, return_type=None):
        """
        Returns a string that can be used as a reference to the entity metadata resource,
        optionally with a specified type parameter added.

        Extracts appropriate local reference, and combines with entity URL path.
        """
        turtle_ref = self.get_entity_turtle_ref(return_type=return_type)
        turtle_url = self.get_src_entity_resource_url(turtle_ref)
        log.debug(
            "get_entity_turtle_url: _entityfile %s, turtle_ref %s, turtle_url %s"%
            (self.curr_typeinfo.entityclass._entityfile, turtle_ref, turtle_url)
            )
        return turtle_url

    def get_entity_list_ref(self, list_name=layout.ENTITY_LIST_FILE, return_type=None):
        """
        Returns a string that can be used as a reference to the entity list data
        relative to the current list URL.
        """
        return make_data_ref(
            self.view.get_request_path(), 
            list_name, 
            return_type
            )

    def context_data(self, entity_label=None):
        """
        Return dictionary of rendering context data available from the 
        elements assembled.

        Values that are added here to the view context are used for view rendering, 
        and are not passed to the entity value mapping process.

        NOTE: values that are needed to be accessible as part of bound_field values 
        must be provided earlier in the form generation process, as elements of the 
        "context_extra_values" dictionary.

        Context values set here do not need to be named in the valuye map used to
        create the view context.
        """
        site_url_parts = urlsplit(self.site._entityurl)
        context = (
            { 'site_label':         self.sitedata["title"]
            , 'title':              self.sitedata["title"]
            , 'heading':            self.sitedata["title"]
            , 'action':             self.action
            , 'coll_id':            self.coll_id
            , 'type_id':            self.type_id
            , 'view_id':            self.view_id
            , 'list_id':            self.list_id
            , 'collection':         self.collection
            , 'info_coll_id':       self.coll_id or layout.SITEDATA_ID
            , "SITE":               site_url_parts.path
            , "HOST":               self.reqhost            
            })
        context.update(self.authorizations)
        if self.collection:
            coll_url_parts = urlsplit(self.collection._entityurl)
            context.update(
                { 'heading':    self.collection[RDFS.CURIE.label]
                , 'coll_label': self.collection[RDFS.CURIE.label]
                , "COLL":       coll_url_parts.path
                , "BASE":       coll_url_parts.path + layout.COLL_BASE_REF
                , "PAGE":       coll_url_parts.path + layout.COLL_PAGE_REF
                })
            context['title'] = "%(coll_label)s"%context
        if self.recordview:
            context.update(
                { 'heading':            self.recordview[RDFS.CURIE.label]
                , 'view_label':         self.recordview[RDFS.CURIE.label]
                , 'edit_view_button':   self.recordview.get(ANNAL.CURIE.open_view, "yes")
                })
            context['title'] = "%(view_label)s - %(coll_label)s"%context
            task_buttons      = self.recordview.get(ANNAL.CURIE.task_buttons, None)
            edit_task_buttons = self.recordview.get(ANNAL.CURIE.edit_task_buttons, task_buttons)
            view_task_buttons = self.recordview.get(ANNAL.CURIE.view_task_buttons, task_buttons)
            self.add_task_button_context('edit_task_buttons', edit_task_buttons, context)
            self.add_task_button_context('view_task_buttons', view_task_buttons, context)
        if self.recordlist:
            context.update(
                { 'heading':                self.recordlist[RDFS.CURIE.label]
                , 'list_label':             self.recordlist[RDFS.CURIE.label]
                , 'entity_list_ref':        self.get_entity_list_ref()
                , 'entity_list_ref_json':   self.get_entity_list_ref(return_type="application/json")
                , 'entity_list_ref_turtle': self.get_entity_list_ref(list_name=layout.ENTITY_LIST_TURTLE)
                })
            context['title'] = "%(list_label)s - %(coll_label)s"%context
        if self.curr_typeinfo:
            context.update(
                { 'entity_data_ref':        self.get_entity_jsonld_url()
                , 'entity_data_ref_json':   self.get_entity_jsonld_url(return_type="application/json")
                , 'entity_turtle_ref':      self.get_entity_turtle_url()
                })
        if entity_label:
            context.update(
                { 'entity_label':       entity_label
                })
            # context['heading'] = "%(entity_label)s - %(view_label)s"%context
            context['title']   = "%(entity_label)s - %(view_label)s - %(coll_label)s"%context
        if hasattr(self.view, 'help') and self.view.help:
            context.update(
                { 'help_filename':  self.view.help
                })
        if hasattr(self.view, 'help_markdown') and self.view.help_markdown:
            substituted_text = apply_substitutions(context, self.view.help_markdown)
            context.update(
                { 'help_markdown':  substituted_text
                })
        if self.info_messages:
            context.update(
                { "info_head":      message.ACTION_COMPLETED
                , "info_message":   "\n".join(self.info_messages)
                })
        if self.error_messages:
            context.update(
                { "error_head":     message.DATA_ERROR
                , "error_message":  "\n".join(self.error_messages)
                })
        return context

    def add_task_button_context(self, task_buttons_name, task_buttons, context):
        """
        Adds context values to a supplied context dictionary corresponding 
        to the supplied task_buttons value(s) from a view description.

        @NOTE: subsequent versions of this function may extract values from
        an identified task description record.
        """
        if isinstance(task_buttons, list):
            context.update(
                { task_buttons_name:
                    [ { 'button_id':    b[ANNAL.CURIE.button_id]
                      , 'button_name':  extract_entity_id(b[ANNAL.CURIE.button_id])
                      , 'button_label': b.get(ANNAL.CURIE.button_label, "@@annal:button_label@@")
                      , 'button_help':  b.get(ANNAL.CURIE.button_help,  "@@annal:button_help@@")
                      } for b in task_buttons
                    ]
                })
        elif task_buttons is not None:
            log.error(
                "DisplayInfo.context_data: Unexpected value for task_buttons: %r"%
                (task_buttons)
                )
        return

    def __str__(self):
        attrs = (
            [ "view"
            , "action"
            , "authorizations"
            , "reqhost"
            , "site"
            , "sitedata"
            , "coll_id"
            , "collection"
            , "type_id"
            , "entitytypeinfo"
            , "list_id"
            , "recordlist"
            , "view_id"
            , "recordview"
            , "entity_id"
            ])
        fields = ["DisplayInfo("]
        for attr in attrs:
            val = getattr(self, attr, None)
            if val is not None:
                fields.append("%s: %r"%(attr, val))
        fields.append(")")
        return ("\n  ".join(fields))

    def __repr__(self):
        return str(self)

if __name__ == "__main__":
    import doctest
    doctest.testmod()

# End.



"""
Utility functions to support entity data testing
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import os
import copy
from collections import OrderedDict

from django.conf                import settings
from django.http                import QueryDict
from django.utils.http          import urlquote, urlunquote
from django.core.urlresolvers   import resolve, reverse

from utils.py3porting           import urljoin

from annalist.util              import valid_id
from annalist.identifiers       import RDF, RDFS, ANNAL
from annalist                   import layout

from annalist.views.form_utils.fieldchoice  import FieldChoice
from annalist.views.fields.render_placement import (
    get_placement_classes
    )

from .tests import (
    TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
    )
from .entity_testutils import (
    collection_dir, 
    site_title, 
    collection_entity_view_url,
    context_field_row
    )
from .entity_testfielddesc       import get_field_description, get_bound_field
from .entity_testentitydata      import entitydata_list_type_url

#   -----------------------------------------------------------------------------
#
#   Directory generating functions
#
#   -----------------------------------------------------------------------------

def recordview_dir(coll_id="testcoll", view_id="testview"):
    return collection_dir(coll_id) + layout.COLL_VIEW_PATH%{'id': view_id} + "/"

#   -----------------------------------------------------------------------------
#
#   URI generating functions
#
#   -----------------------------------------------------------------------------

def recordview_coll_url(site, coll_id="testcoll", view_id="testview"):
    return urljoin(
        site._entityurl,
        layout.SITE_COLL_PATH%{'id': coll_id} + "/" + 
        layout.COLL_VIEW_PATH%{'id': view_id} + "/"
        )

def recordview_url(coll_id, view_id):
    """
    URI for record view description data; also view using default entity view
    """
    if not valid_id(view_id):
        view_id = "___"
    return collection_entity_view_url(coll_id=coll_id, type_id="_view", entity_id=view_id)

def recordview_edit_url(action=None, coll_id=None, view_id=None):
    """
    URI for record view description editing view
    """
    viewname = ( 
        'AnnalistEntityDataView'        if action == "view"   else
        'AnnalistEntityNewView'         if action == "new"    else
        'AnnalistEntityEditView'        if action == "copy"   else
        'AnnalistEntityEditView'        if action == "edit"   else
        'AnnalistRecordViewDeleteView'  if action == "delete" else
        'unknown'
        )
    kwargs = {'coll_id': coll_id}
    if action != "delete":
        kwargs.update({'action': action, 'type_id': "_view", 'view_id': "View_view"})
    if view_id:
        if valid_id(view_id):
            kwargs.update({'entity_id': view_id})
        else:
            kwargs.update({'entity_id': "___"})
    return reverse(viewname, kwargs=kwargs)

#   -----------------------------------------------------------------------------
#
#   ----- RecordView data for default view
#
#   -----------------------------------------------------------------------------

def recordview_value_keys(view_uri=False, view_entity_type=True):
    keys = set(
        [ 'annal:id', 'annal:type_id'
        , 'annal:type'
        , 'annal:url'
        , 'rdfs:label', 'rdfs:comment'
        , 'annal:open_view'
        , 'annal:view_fields'
        ])
    if view_uri:
        keys.add('annal:uri')
    if view_entity_type or (view_entity_type == ""):
        keys.add('annal:view_entity_type')
    return keys

def recordview_load_keys(view_uri=False, view_entity_type=True):
    return (
        recordview_value_keys(view_uri=view_uri, view_entity_type=view_entity_type) | 
        {"@id", '@type', '@context'}
        )

def recordview_create_values(
        coll_id="testcoll", view_id="testview", update="RecordView", view_uri=None, 
        view_entity_type="annal:Test_default",
        num_fields=4, field3_placement="small:0,12",
        extra_field=None, extra_field_uri=None
        ):
    """
    Entity values used when creating a record view entity
    """
    view_values = (
        { 'annal:type':         "annal:View"
        , 'rdfs:label':         "%s %s/%s"%(update, coll_id, view_id)
        , 'rdfs:comment':       "%s help for %s in collection %s"%(update, view_id, coll_id)
        , 'annal:view_entity_type':  view_entity_type
        , 'annal:open_view':    True
        , 'annal:view_fields':
          [ { 'annal:field_id':         layout.FIELD_TYPEID+"/Entity_id"
            , 'annal:field_placement':  "small:0,12;medium:0,6"
            }
          , { 'annal:field_id':         layout.FIELD_TYPEID+"/Entity_type"
            , 'annal:field_placement':  "small:0,12;medium:6,6"
            }
          , { 'annal:field_id':         layout.FIELD_TYPEID+"/Entity_label"
            , 'annal:field_placement':  "small:0,12"
            }
          , { 'annal:field_id':         layout.FIELD_TYPEID+"/Entity_comment"
            # , 'annal:field_placement':  field3_placement
            }
          ]
        })
    if view_uri:
        view_values['annal:uri'] = view_uri
    if field3_placement:
        view_values['annal:view_fields'][3]['annal:field_placement'] = field3_placement
    if extra_field:
        efd = (
            { 'annal:field_id':         extra_field
            , 'annal:field_placement':  "small:0,12"
            })
        if extra_field_uri:
            efd['annal:property_uri'] = extra_field_uri
        view_values['annal:view_fields'].append(efd)
    if num_fields == 0:
        view_values['annal:view_fields'] = []
    return view_values

def recordview_values(
        coll_id="testcoll", view_id="testtype", view_uri=None, 
        update="RecordView", hosturi=TestHostUri, 
        view_entity_type="annal:Test_default",
        num_fields=4, field3_placement="small:0,12",
        extra_field=None, extra_field_uri=None):
    d = recordview_create_values(
        coll_id, view_id, update=update, view_uri=view_uri,
        view_entity_type=view_entity_type,
        num_fields=num_fields, field3_placement=field3_placement,
        extra_field=extra_field, extra_field_uri=extra_field_uri
        ).copy()
    view_url = recordview_url(coll_id, view_id)
    d.update(
        { 'annal:id':       view_id
        , 'annal:type_id':  "_view"
        , 'annal:url':      view_url
        })
    return d

def recordview_read_values(
        coll_id="testcoll", view_id="testview", view_uri=None, 
        view_entity_type="annal:Test_default",
        update="RecordView", hosturi=TestHostUri):
    d = recordview_values(
        coll_id, view_id, view_uri=view_uri, 
        view_entity_type=view_entity_type,
        update=update, hosturi=hosturi
        ).copy()
    d.update(
        { '@id':            layout.COLL_BASE_VIEW_REF%{'id': view_id}
        , '@type':          ["annal:View"]
        , '@context':       [{"@base": "../../"}, "../../coll_context.jsonld"]
        })
    return d

def recordview_values_add_field(view_data, 
    field_id="Entity_comment", 
    field_property_uri="rdfs:comment",
    field_placement="small:0,12"
    ):
    view_data['annal:view_fields'].append(
        { 'annal:field_id':         field_id
        , 'annal:property_uri':     field_property_uri
        , 'annal:field_placement':  field_placement
        })
    return view_data

#   -----------------------------------------------------------------------------
#
#   ----- Data in recordview view for default entity data
#
#   -----------------------------------------------------------------------------

default_view_fields_list = (
    [ { ANNAL.CURIE.field_id:           layout.FIELD_TYPEID+"/Entity_id"
      # , ANNAL.CURIE.property_uri:       ANNAL.CURIE.id
      , ANNAL.CURIE.field_placement:    "small:0,12;medium:0,6"
      }
    , { ANNAL.CURIE.field_id:           layout.FIELD_TYPEID+"/Entity_type"
      # , ANNAL.CURIE.property_uri:       ANNAL.CURIE.type_id
      , ANNAL.CURIE.field_placement:    "small:0,12;medium:6,6"
      }
    , { ANNAL.CURIE.field_id:           layout.FIELD_TYPEID+"/Entity_label"
      # , ANNAL.CURIE.property_uri:       RDFS.CURIE.label
      , ANNAL.CURIE.field_placement:    "small:0,12"
      }
    , { ANNAL.CURIE.field_id:           layout.FIELD_TYPEID+"/Entity_comment"
      # , ANNAL.CURIE.property_uri:       RDFS.CURIE.comment
      , ANNAL.CURIE.field_placement:    "small:0,12"
      }
    ])

view_view_fields_list = (
    [
      { ANNAL.CURIE.field_id:           layout.FIELD_TYPEID+'/View_id'
      , ANNAL.CURIE.property_uri:       ANNAL.CURIE.id
      , ANNAL.CURIE.field_placement:    "small:0,12;medium:0,6"
      }
    , { ANNAL.CURIE.field_id:           layout.FIELD_TYPEID+"/View_label"
      , ANNAL.CURIE.property_uri:       RDFS.CURIE.label
      , ANNAL.CURIE.field_placement:    "small:0,12"
      }
    , { ANNAL.CURIE.field_id:           layout.FIELD_TYPEID+"/View_comment"
      , ANNAL.CURIE.property_uri:       RDFS.CURIE.comment
      , ANNAL.CURIE.field_placement:    "small:0,12"
      }
    , { ANNAL.CURIE.field_id:           layout.FIELD_TYPEID+"/View_entity_type"
      , ANNAL.CURIE.property_uri:       ANNAL.CURIE.view_entity_type
      , ANNAL.CURIE.field_placement:    "small:0,12"
      }
    , { ANNAL.CURIE.field_id:           layout.FIELD_TYPEID+"/View_edit_view"
      , ANNAL.CURIE.property_uri:       ANNAL.CURIE.open_view
      , ANNAL.CURIE.field_placement:    "small:0,12;medium:0,6"
      }
    , { ANNAL.CURIE.field_id:           layout.FIELD_TYPEID+"/View_fields"
      , ANNAL.CURIE.property_uri:       ANNAL.CURIE.view_fields
      , ANNAL.CURIE.field_placement:    "small:0,12"
      }
    ])

# Field list within a view_fields field of a view display
# (this is confusingly, but unavoidably, self-referential)
group_field_list = (
    [ { ANNAL.CURIE.field_id:           "_field/View_field_sel"
      , ANNAL.CURIE.property_uri:       ANNAL.CURIE.field_id
      , ANNAL.CURIE.field_placement:    "small:0,12;medium:0,4"
      }
    , { ANNAL.CURIE.field_id:           "_field/View_field_property"
      , ANNAL.CURIE.property_uri:       ANNAL.CURIE.property_uri
      , ANNAL.CURIE.field_placement:    "small:0,12;medium:4,4"
      }
    , { ANNAL.CURIE.field_id:           "_field/View_field_placement"
      , ANNAL.CURIE.property_uri:       ANNAL.CURIE.field_placement
      , ANNAL.CURIE.field_placement:    "small:0,12;medium:8,4"
      }
    ])

def view_view_context_data(
        coll_id="testcoll", view_id="", orig_id=None,
        action=None, 
        view_uri=None,
        view_label=None,
        view_descr=None,
        view_entity_type="annal:Default_type",
        view_edit_view=True,
        view_fields=None, 
        add_field=None, remove_field=None, move_up=None, move_down=None,
        field_choices=None,
        update="RecordView",
        continuation_url=None
    ):
    if view_label is None:
        #@@TODO: use same format for both; change form data too
        if view_id:
            view_label = "%s %s/%s"%(update, coll_id, view_id)
        else:
            view_label = "%s data ... (%s/)"%(update, coll_id)
    # Target record fields listed in the view description
    if view_fields is None:
        view_fields = default_view_fields_list
    if continuation_url is None:
        continuation_url = entitydata_list_type_url(coll_id, layout.VIEW_TYPEID)
    view_heading = "View definition"
    view_title   = (
        "%s - %s - Collection %s"%(view_label, view_heading, coll_id) if view_label
        else
        "%s - Collection %s"%(view_heading, coll_id)
        )
    view_fields = copy.deepcopy(view_fields) # Prevents testing Heisenbugs
    bound_view_fields = get_bound_field("View_fields", view_fields)
    if field_choices is not None:
        # Add field choices to field definitions...
        bound_view_fields['description']['group_field_list'] = group_field_list
        group_field_descs = (
            [ get_field_description(f[ANNAL.CURIE.field_id]) 
              for f in group_field_list 
            ])
        fcs = OrderedDict()
        for fid in field_choices:
            fref      = layout.FIELD_TYPEID + "/" + fid
            fdesc     = get_field_description(fid)
            fc        = FieldChoice(fref, 
                label=fdesc['field_label'],
                link=TestBasePath+"/c/"+coll_id+"/d/"+fref+"/"
                )
            fcs[fref] = fc
        group_field_descs[0]['field_choices'] = fcs
        bound_view_fields['description']['group_field_descs'] = group_field_descs
    context_dict = (
        { 'title':              view_title
        , 'heading':            view_heading
        , 'coll_id':            coll_id
        , 'type_id':            layout.VIEW_TYPEID
        , 'orig_id':            orig_id
        , 'view_id':            'View_view'
        , 'entity_id':          view_id or ""
        , 'orig_id':            orig_id
        , 'orig_type':          layout.VIEW_TYPEID
        , 'record_type':        "annal:View"
        , 'continuation_url':   continuation_url
        , 'fields':
          [ context_field_row(
              get_bound_field("View_id",           view_id),                # 0 (0,0)
              )
          , context_field_row(
              get_bound_field("View_label",        view_label)              # 1 (1,0)
              )
          , context_field_row(
              get_bound_field("View_comment",      view_descr)              # 2 (2,0)
              )
          , context_field_row(
              get_bound_field("View_entity_type",  view_entity_type)        # 3 (3,0)
              )
          , context_field_row(
              get_bound_field("View_edit_view",    view_edit_view)          # 4 (4,0)
              )
          , bound_view_fields  # 5 (5, 0)
          ]
        })
    if view_uri:
        context_dict['entity_uri']  = view_uri
    if action:  
        context_dict['action']      = action
    if add_field:
        context_dict['fields'][5]['field_value'].append(
            { 'annal:field_id':         None
            , 'annal:property_uri':     None
            , 'annal:field_placement':  None
            })
    if remove_field:
        context_dict['fields'][5]['field_value'][3:4] = []
    if move_up:
        assert move_up == [2,3]
        fields = context_dict['fields'][5]['field_value']
        context_dict['fields'][5]['field_value'] = [ fields[i]  for i in [0,2,3,1] ]
    if move_down:
        assert move_down == [1]
        fields = context_dict['fields'][5]['field_value']
        context_dict['fields'][5]['field_value'] = [ fields[i]  for i in [0,2,1,3] ]
    return context_dict

def view_view_form_data(
        coll_id="testcoll", orig_coll=None,
        view_id="", orig_id=None, 
        action=None, cancel=None, 
        view_entity_type="annal:View",
        field3_placement="small:0,12",
        extra_field=None,       # Extra field id for some tests (e.g. dup property uri)
        extra_field_uri=None,   # Extra field property URI to add in
        add_field=None,         # True for add field option
        remove_fields=None,     # List of field numbers to remove (as strings)
        move_up_fields=None,    # List of field numbers to move up
        move_down_fields=None,  # List of field numbers to move down
        new_enum=None,          # Id of new-value button enumerated value field
        update="RecordView"
        ):
    form_data_dict = (
        { 'View_label':         '%s data ... (%s/%s)'%(update, coll_id, view_id)
        , 'View_comment':       '%s description ... (%s/%s)'%(update, coll_id, view_id)
        , 'View_entity_type':   view_entity_type
        , 'View_edit_view':     "Yes"
        , 'orig_id':            'orig_view_id'
        , 'record_type':        'annal:View'
        , 'continuation_url':   entitydata_list_type_url(coll_id, "_view")
        # View fields
        , 'View_fields__0__View_field_sel':          layout.FIELD_TYPEID+"/Entity_id"
        , 'View_fields__0__View_field_placement':    "small:0,12;medium:0,6"
        , 'View_fields__1__View_field_sel':          layout.FIELD_TYPEID+"/Entity_type"
        , 'View_fields__1__View_field_placement':    "small:0,12;medium:6,6"
        , 'View_fields__2__View_field_sel':          layout.FIELD_TYPEID+"/Entity_label"
        , 'View_fields__2__View_field_property':     "rdfs:label"
        , 'View_fields__2__View_field_placement':    "small:0,12"
        , 'View_fields__3__View_field_sel':          layout.FIELD_TYPEID+"/Entity_comment"
        , 'View_fields__3__View_field_property':     "rdfs:comment"
        , 'View_fields__3__View_field_placement':    field3_placement
        })
    if extra_field:
        # Insert extra field with supplied Id
        form_data_dict['View_fields__4__View_field_sel']       = layout.FIELD_TYPEID+"/"+extra_field
        form_data_dict['View_fields__4__View_field_placement'] = "small:0,12"
        if extra_field_uri:
            form_data_dict['View_fields__4__View_field_property'] = extra_field_uri
    if view_id is not None:
        form_data_dict['entity_id']     = view_id
    if view_id:
        form_data_dict['entity_id']     = view_id
        form_data_dict['View_label']    = '%s %s/%s'%(update, coll_id, view_id)
        form_data_dict['View_comment']  = '%s help for %s in collection %s'%(update, view_id, coll_id)
        form_data_dict['View_uri']      = TestBaseUri + "/c/%s/d/_view/%s/"%(coll_id, view_id)
        form_data_dict['orig_id']       = view_id
        form_data_dict['orig_type']     = "_view"
        form_data_dict['orig_coll']     = coll_id
    if orig_id:
        form_data_dict['orig_id']       = orig_id
    if orig_coll:
        form_data_dict['orig_coll']     = orig_coll
    if action:
        form_data_dict['action']        = action
    if cancel:
        form_data_dict['cancel']        = "Cancel"
    elif add_field:
        form_data_dict['View_fields__add'] = "Add field"
    elif remove_fields:
        form_data_dict['View_fields__remove'] = "Remove field"
        if remove_fields != "no-selection":
            form_data_dict['View_fields__select_fields'] = remove_fields
    elif move_up_fields:
        form_data_dict['View_fields__up'] = "Move up"
        if move_up_fields != "no-selection":
            form_data_dict['View_fields__select_fields'] = move_up_fields
    elif move_down_fields:
        form_data_dict['View_fields__down'] = "Move down"
        if move_down_fields != "no-selection":
            form_data_dict['View_fields__select_fields'] = move_down_fields
    elif new_enum:
        form_data_dict[new_enum]        = new_enum
    else:
        form_data_dict['save']          = 'Save'
    return form_data_dict

#   -----------------------------------------------------------------------------
#
#   ----- Recordview delete confirmation form data
#
#   -----------------------------------------------------------------------------

def recordview_delete_confirm_form_data(view_id=None):
    return (
        { 'viewlist':    view_id,
          'view_delete': 'Delete'
        })

# End.



"""
Test module for annalist-manager site data management commands
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2018, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import sys
import os

from utils.py3porting           import StringIO
from utils.StdoutContext        import SwitchStdout, SwitchStderr

import annalist
from annalist.util              import replacetree, removetree

from annalist_manager.tests     import get_source_root
from annalist_manager.tests     import test_annalist_base
from annalist_manager.am_main   import runCommand

#   -----------------------------------------------------------------------------
#
#   Tests
#
#   -----------------------------------------------------------------------------

class AnnalistManagerSiteTest(test_annalist_base.AnnalistManagerTestBase):

    @classmethod
    def setUpTestData(cls):
        cls.setup_annalist_manager_test()
        return

    def setUp(self):
        if os.path.isdir(self.sitehome):
            removetree(self.sitehome)
        return

    def tearDown(self):
        return

    #   -----------------------------------------------------------------------------
    #   Tests
    #   -----------------------------------------------------------------------------

    def test_createsitedata(self):
        stdoutbuf  = StringIO()
        with SwitchStdout(stdoutbuf):
            runCommand(self.userhome, self.userconfig, 
                ["annalist-manager", "createsitedata", "--config=runtests"]
                )
        stdoutbuf.seek(0)
        stdoutlines = stdoutbuf.read().split("\n")
        self.assertEqual(stdoutlines[0], "Initializing Annalist site in "+self.sitehome)
        siteexists = os.path.isdir(self.sitehome)
        collexists = os.path.isfile(os.path.join(self.sitehome, "c/_annalist_site/d/coll_meta.jsonld"))
        self.assertTrue(siteexists, "Annalist site directory exists?")
        self.assertTrue(collexists, "Annalist site collection metadata exists?")
        return

    def test_updatesitedata(self):
        stdoutbuf  = StringIO()
        with SwitchStdout(stdoutbuf):
            runCommand(self.userhome, self.userconfig, 
                ["annalist-manager", "createsitedata", "--config=runtests"]
                )
        stdoutbuf.seek(0)
        stdoutbuf  = StringIO()
        with SwitchStdout(stdoutbuf):
            runCommand(self.userhome, self.userconfig, 
                ["annalist-manager", "updatesitedata", "--config=runtests"]
                )
        stdoutbuf.seek(0)
        stdoutlines = stdoutbuf.read().split("\n")
        self.assertEqual(stdoutlines[0], "Copy Annalist site data")
        siteexists = os.path.isdir(self.sitehome)
        collexists = os.path.isfile(os.path.join(self.sitehome, "c/_annalist_site/d/coll_meta.jsonld"))
        self.assertTrue(siteexists, "Annalist site directory exists?")
        self.assertTrue(collexists, "Annalist site collection metadata exists?")
        return

# End.



"""
Test module for annalist-manager user management
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2018, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import sys
import os
import json

from utils.py3porting           import StringIO
from utils.StdoutContext        import SwitchStdout, SwitchStderr, SwitchStdin

import annalist
from annalist.util              import replacetree, removetree

from annalist_manager.tests     import get_source_root
from annalist_manager.tests     import test_annalist_base
from annalist_manager.am_main   import runCommand

#   -----------------------------------------------------------------------------
#
#   Tests
#
#   -----------------------------------------------------------------------------

class AnnalistManagerUserTest(test_annalist_base.AnnalistManagerTestBase):

    @classmethod
    def setUpTestData(cls):
        cls.setup_annalist_manager_test()
        cls.init_site_database()
        cls.create_site_data()
        return

    def setUp(self):
        return

    def tearDown(self):
        return

    #   -----------------------------------------------------------------------------
    #   Helpers
    #   -----------------------------------------------------------------------------

    def createuser(self, username):
        stderrbuf  = StringIO()
        with SwitchStderr(stderrbuf):
            stdoutbuf  = StringIO()
            with SwitchStdout(stdoutbuf):
                stdinbuf = StringIO("password\npassword\n")
                with SwitchStdin(stdinbuf):
                    runCommand(self.userhome, self.userconfig, 
                        [ "annalist-manager", "createlocal"
                        , username, username+"@example.org", "Test", username
                        , "--config=runtests"
                        ])
        return

    def userdir(self, username):
        usertypedir = os.path.join(self.sitehome, "c/_annalist_site/d/_user/")
        return os.path.join(usertypedir, username)

    def check_user_permissions(self, username, permissions):
        userfile = os.path.join(self.userdir(username), "user_meta.jsonld")
        with open(userfile) as f:
            userdata = json.load(f)
        self.assertEqual(userdata["annal:user_permissions"], permissions.split())
        return

    #   -----------------------------------------------------------------------------
    #   Tests
    #   -----------------------------------------------------------------------------

    def test_createlocaluser(self):
        stderrbuf  = StringIO()
        with SwitchStderr(stderrbuf):
            stdoutbuf  = StringIO()
            with SwitchStdout(stdoutbuf):
                stdinbuf = StringIO("password\npassword\n")
                with SwitchStdin(stdinbuf):
                    runCommand(self.userhome, self.userconfig, 
                        [ "annalist-manager", "createlocaluser"
                        , "testlocaluser", "testlocaluser@example.org", "Test", "LocalUser"
                        , "--config=runtests"
                        ])
        userexists = os.path.isdir(self.userdir("testlocaluser"))
        self.assertTrue(userexists, "testlocaluser created OK?")
        return

    def test_createadminuser(self):
        stderrbuf  = StringIO()
        with SwitchStderr(stderrbuf):
            stdoutbuf  = StringIO()
            with SwitchStdout(stdoutbuf):
                stdinbuf = StringIO("password\npassword\n")
                with SwitchStdin(stdinbuf):
                    runCommand(self.userhome, self.userconfig, 
                        [ "annalist-manager", "createadminuser"
                        , "testadminuser", "testadminuser@example.org", "Test", "LocalUser"
                        , "--config=runtests"
                        ])
        userexists = os.path.isdir(self.userdir("testadminuser"))
        self.assertTrue(userexists, "testadminuser created OK?")
        return

    def test_defaultadminuser(self):
        stderrbuf  = StringIO()
        with SwitchStderr(stderrbuf):
            stdoutbuf  = StringIO()
            with SwitchStdout(stdoutbuf):
                stdinbuf = StringIO("password\npassword\n")
                with SwitchStdin(stdinbuf):
                    runCommand(self.userhome, self.userconfig, 
                        [ "annalist-manager", "defaultadminuser"
                        , "--config=runtests"
                        ])
        userexists = os.path.isdir(self.userdir("admin"))
        self.assertTrue(userexists, "admin created OK?")
        return

    def test_updateadminuser(self):
        self.createuser("testupdateuser")
        userexists = os.path.isdir(self.userdir("testupdateuser"))
        self.assertTrue(userexists, "testupdateuser created OK?")
        stderrbuf  = StringIO()
        with SwitchStderr(stderrbuf):
            stdoutbuf  = StringIO()
            with SwitchStdout(stdoutbuf):
                runCommand(self.userhome, self.userconfig, 
                    [ "annalist-manager", "updateadminuser"
                    , "testupdateuser"
                    , "--config=runtests"
                    ])
        self.check_user_permissions("testupdateuser", "VIEW CREATE UPDATE DELETE CONFIG ADMIN")
        return

    def test_setuserpermissions(self):
        self.createuser("testupdateuser")
        userexists = os.path.isdir(self.userdir("testupdateuser"))
        self.assertTrue(userexists, "testupdateuser created OK?")
        stderrbuf  = StringIO()
        with SwitchStderr(stderrbuf):
            stdoutbuf  = StringIO()
            with SwitchStdout(stdoutbuf):
                runCommand(self.userhome, self.userconfig, 
                    [ "annalist-manager", "setuserpermissions"
                    , "testupdateuser"
                    , "CREATE_COLLECTION DELETE_COLLECTION VIEW CREATE UPDATE DELETE CONFIG"
                    , "--config=runtests"
                    ])
        self.check_user_permissions(
            "testupdateuser", 
            "CREATE_COLLECTION DELETE_COLLECTION VIEW CREATE UPDATE DELETE CONFIG"
            )
        return

    def test_setdefaultpermissions(self):
        stderrbuf  = StringIO()
        with SwitchStderr(stderrbuf):
            stdoutbuf  = StringIO()
            with SwitchStdout(stdoutbuf):
                runCommand(self.userhome, self.userconfig, 
                    [ "annalist-manager", "setdefaultpermissions"
                    , "CREATE_COLLECTION VIEW CREATE"
                    , "--config=runtests"
                    ])
        self.check_user_permissions(
            "_default_user_perms", 
            "CREATE_COLLECTION VIEW CREATE"
            )
        return

    def test_setpublicpermissions(self):
        stderrbuf  = StringIO()
        with SwitchStderr(stderrbuf):
            stdoutbuf  = StringIO()
            with SwitchStdout(stdoutbuf):
                runCommand(self.userhome, self.userconfig, 
                    [ "annalist-manager", "setpublicpermissions"
                    , "VIEW CREATE"
                    , "--config=runtests"
                    ])
        self.check_user_permissions(
            "_unknown_user_perms", 
            "VIEW CREATE"
            )
        return

    def test_deleteuser(self):
        self.createuser("testdeleteuser")
        userexists = os.path.isdir(self.userdir("testdeleteuser"))
        self.assertTrue(userexists, "testdeleteuser created OK?")
        stderrbuf  = StringIO()
        with SwitchStderr(stderrbuf):
            stdoutbuf  = StringIO()
            with SwitchStdout(stdoutbuf):
                runCommand(self.userhome, self.userconfig, 
                    [ "annalist-manager", "deleteuser"
                    , "testdeleteuser"
                    , "--config=runtests"
                    ])
        # print("stderrbuf:")
        # stderrbuf.seek(0)
        # print(stderrbuf.read())
        userexists = os.path.isdir(self.userdir("testdeleteuser"))
        self.assertFalse(userexists, "testdeleteuser deleted OK?")
        return

# End.



from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

"""
Tests for generic page elements not covered by other tests
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2018, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import unittest

import logging
log = logging.getLogger(__name__)

from django.core.urlresolvers   import resolve, reverse
from django.test.client         import Client

from annalist                   import layout
from annalist.models.site       import Site
from annalist.models.collection import Collection
from annalist.views.serverlog   import ServerLogView

from .AnnalistTestCase import AnnalistTestCase
from .tests import (
    TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
    )
from .init_tests import (
    init_annalist_test_site,
    init_annalist_test_coll,
    resetSitedata
    )
from .entity_testutils import (
    collection_create_values,
    create_test_user, create_user_permissions, 
    )

class ServerLogTest(AnnalistTestCase):
    """
    Tests for user profile page
    """

    def setUp(self):
        init_annalist_test_site()
        self.testsite = Site(TestBaseUri, TestBaseDir)
        return

    def tearDown(self):
        return

    @classmethod
    def setUpClass(cls):
        super(ServerLogTest, cls).setUpClass()
        return

    @classmethod
    def tearDownClass(cls):
        super(ServerLogTest, cls).tearDownClass()
        return

    def test_ServerLogTest(self):
        self.assertEqual(ServerLogView.__name__, "ServerLogView", "Check ServerLogView class name")
        return

#   -----------------------------------------------------------------------------
#
#   ServerLogView tests
#
#   -----------------------------------------------------------------------------

class UserServerLogViewTest(AnnalistTestCase):
    """
    Tests for ServerLog views
    """

    def setUp(self):
        init_annalist_test_site()
        self.testsite = Site(TestBaseUri, TestBaseDir)
        self.testcoll = Collection.create(
            self.testsite, "testcoll", collection_create_values("testcoll")
            )
        self.sitecoll = Collection.load(self.testsite, layout.SITEDATA_ID)
        self.client   = Client(HTTP_HOST=TestHost)
        self.uri      = reverse("AnnalistServerLogView")
        self.homeuri  = reverse("AnnalistHomeView")
        return

    def tearDown(self):
        return

    def test_get(self):
        create_test_user(
            self.sitecoll, "adminuser", "adminpassword",
            user_permissions=["VIEW", "CREATE", "UPDATE", "DELETE", "CONFIG", "ADMIN"]
            )
        loggedin = self.client.login(username="adminuser", password="adminpassword")
        self.assertTrue(loggedin)
        r = self.client.get(self.uri)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        site_title = "Annalist data notebook test site"
        self.assertEqual(r.context['title'], site_title)
        self.assertIn("serverlog", r.context)
        self.assertContains(r, site_title)
        return

    def test_get_no_admin(self):
        create_test_user(
            self.testcoll, "testuser", "testpassword",
            user_permissions=["VIEW", "CREATE", "UPDATE", "DELETE", "CONFIG"]
            )
        loggedin = self.client.login(username="testuser", password="testpassword")
        self.assertTrue(loggedin)
        r = self.client.get(self.uri)
        self.assertEqual(r.status_code,   403)
        self.assertEqual(r.reason_phrase, "Forbidden")
        err_head    = "403: Forbidden"
        err_message = "No ADMIN access permission for resource http://test.example.com/testsite/serverlog/"
        self.assertEqual(r.context["message"], err_message)
        self.assertContains(r, err_head,    status_code=403)
        self.assertContains(r, err_message, status_code=403)
        return

    def test_get_no_login(self):
        self.client.logout()
        r = self.client.get(self.uri)
        self.assertEqual(r.status_code,   401)
        self.assertEqual(r.reason_phrase, "Unauthorized")
        err_head    = "401: Unauthorized"
        err_message = "Resource http://test.example.com/testsite/serverlog/ requires authentication for ADMIN access"
        self.assertEqual(r.context["message"], err_message)
        self.assertContains(r, err_head,    status_code=401)
        self.assertContains(r, err_message, status_code=401)
        return

# End.



"""
Entity resource access (imported and uploaded files).

This module contains functions to help access resource data that is 
attached to an entity (or collection).

It contains logic for accessing raw JSON-LD data, or accessing that data converted to 
some other format (e.g. Turtle).
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2017, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import sys
import os
import json

from rdflib                             import Graph, URIRef, Literal

# Used by `json_resource_file` below.
# See: https://stackoverflow.com/questions/51981089
from utils.py3porting import BytesIO, StringIO, write_bytes

from annalist                           import message
from annalist                           import layout

from annalist.models.entitytypeinfo     import EntityTypeInfo

# Resource info data for built-in entity data

site_fixed_json_resources = (
    [ { "resource_name": layout.SITE_META_FILE,     "resource_dir": layout.SITE_META_REF, 
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.SITE_CONTEXT_FILE,  "resource_dir": layout.SITE_META_REF, 
                                                    "resource_type": "application/ld+json" }
    ])

collection_fixed_json_resources = (
    [ { "resource_name": layout.COLL_META_FILE,     "resource_dir": layout.COLL_BASE_DIR, 
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.COLL_PROV_FILE,     "resource_dir": layout.COLL_BASE_DIR, 
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.COLL_CONTEXT_FILE,  "resource_dir": layout.COLL_BASE_DIR, 
                                                    "resource_type": "application/ld+json" }
    ])

entity_fixed_json_resources = (
    [ { "resource_name": layout.COLL_META_FILE,     "resource_dir": layout.COLL_BASE_DIR, 
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.COLL_PROV_FILE,     "resource_dir": layout.COLL_BASE_DIR, 
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.TYPE_META_FILE,     "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.TYPE_PROV_FILE,     "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.LIST_META_FILE,     "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.LIST_PROV_FILE,     "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.VIEW_META_FILE,     "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.VIEW_PROV_FILE,     "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.GROUP_META_FILE,    "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.GROUP_PROV_FILE,    "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.FIELD_META_FILE,    "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.FIELD_PROV_FILE,    "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.VOCAB_META_FILE,    "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.VOCAB_PROV_FILE,    "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.USER_META_FILE,     "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.USER_PROV_FILE,     "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.INFO_META_FILE,     "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.INFO_PROV_FILE,     "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.ENUM_META_FILE,     "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.ENUM_PROV_FILE,     "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.TYPEDATA_META_FILE, "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.ENTITY_DATA_FILE,   "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    , { "resource_name": layout.ENTITY_PROV_FILE,   "resource_dir": ".",
                                                    "resource_type": "application/ld+json" }
    ])

entity_list_json_resources = (
    # [ { "resource_name": layout.COLL_META_FILE,     "resource_dir": layout.COLL_BASE_DIR, 
    #                                                 "resource_type": "application/ld+json" }
    # , { "resource_name": layout.COLL_PROV_FILE,     "resource_dir": layout.COLL_BASE_DIR, 
    #                                                 "resource_type": "application/ld+json" }
    # , { "resource_name": layout.COLL_CONTEXT_FILE,  "resource_dir": layout.COLL_BASE_DIR, 
    #                                                 "resource_type": "application/ld+json" }
    [ { "resource_name": layout.ENTITY_LIST_FILE,   "resource_dir": ".", 
                                                    "resource_type": "application/ld+json" }
    ])

# Resource access functions

def entity_resource_file(entity, resource_info):
    """
    Return a file object that reads out the content of a resource attached to a specified entity.
    """
    return entity.resource_file(resource_info["resource_path"])

def json_resource_file(baseurl, jsondata, resource_info):
    """
    Return a file object that reads out a JSON version of the supplied entity values data. 

    baseurl         base URL for resolving relative URI references.
                    (Unused except for diagnostic purposes.)
    jsondata        is the data to be formatted and returned.
    resource_info   is a dictionary of values about the resource to be serialized.
                    (Unused except for diagnostic purposes.)
    """
    response_file = StringIO()
    json.dump(jsondata, response_file, indent=2, separators=(',', ': '), sort_keys=True)
    response_file.seek(0)
    return response_file

def turtle_resource_file(baseurl, jsondata, resource_info):
    """
    Return a file object that reads out a Turtle version of the supplied 
    entity values data.  The file object returns a byte stream, as this is
    what rdflib expects.

    baseurl         base URL for resolving relative URI references.
                    (Unused except for diagnostic purposes.)
    jsondata        is the data to be formatted and returned.
    resource_info   is a dictionary of values about the resource to be serialized.
                    (Unused except for diagnostic purposes.)
    """
    # NOTE:   under Python 2, "BytesIO" is implemented by "StringIO", which does
    #         not handle well a combination of str and unicode values, and may 
    #         raise an exception if the Turtle data contains non-ASCII characters.
    #         The problem manifests when an error occurs, and manifests as a 500 
    #         server error response.
    #
    #         On reflection, I think the prioblem arises because the `message.*`
    #         values are unicode (per `from __future__ import unicode_literals`),
    #         and are getting joined with UTF-encoded bytestring values, which
    #         results in the error noted.
    #
    #         The fix here is to encode everything as bytes before writing.
    jsondata_file = json_resource_file(baseurl, jsondata, resource_info)
    response_file = BytesIO()
    g = Graph()
    try:
        g = g.parse(source=jsondata_file, publicID=baseurl, format="json-ld")
    except Exception as e:
        reason = str(e)
        log.warning(message.JSONLD_PARSE_ERROR)
        log.info(reason)
        log.info("baseurl %s, resourceinfo %r"%(baseurl, resource_info))
        write_bytes(response_file, "\n\n***** ERROR ****\n")
        write_bytes(response_file, "%s"%message.JSONLD_PARSE_ERROR)
        write_bytes(response_file, "\n%s:\n"%message.JSONLD_PARSE_REASON)
        write_bytes(response_file, reason)
        write_bytes(response_file, "\n\n")
    try:
        g.serialize(destination=response_file, format='turtle', indent=4)
    except Exception as e:
        reason = str(e)
        log.warning(message.TURTLE_SERIALIZE_ERROR)
        log.info(reason)
        write_bytes(response_file, "\n\n***** ERROR ****\n")
        write_bytes(response_file, "%s"%message.TURTLE_SERIALIZE_ERROR)
        write_bytes(response_file, "\n%s:\n"%message.TURTLE_SERIALIZE_REASON)
        write_bytes(response_file, reason)
        write_bytes(response_file, "\n\n")
    response_file.seek(0)
    return response_file

def make_turtle_resource_info(json_resource):
    """
    Return Turtle resource description for fixed JSON resource
    """
    turtle_resource = (
        { "resource_name":      json_resource["resource_name"][0:-7]+".ttl"
        , "resource_dir":       json_resource["resource_dir"]
        , "resource_type":      "text/turtle"
        , "resource_access":    turtle_resource_file
        })
    return turtle_resource

def find_fixed_resource(fixed_json_resources, resource_ref):
    """
    Return a description for the indicated fixed (built-in) resource from 
    a supplied table, or None

    resource_ref    is the local name of the desired resource relative to the base
                    location of the entity (or collection or site data) to which it belongs.

    The description returned is a dictionary with the following keys:
        resource_name:      filename of resource (i.e. part of URI path after final "/")
        resource_dir:       directory of resource (i.e. part of URI path up to final "/")
        resource_path:      a file or URI path to the resource data relative to the
                            URI of the entity to which it belongs.
        resource_type:      content-type of resource
        resource_access:    optional: if present, specifies a function that returns an 
                            alternative representation of a JSON-LD data resource.
                            (e.g. see `turtle_resource_file`)
    """
    # log.debug("CollectionResourceAccess.find_resource %s/d/%s"%(coll.get_id(), resource_ref))
    for fj in fixed_json_resources:
        if fj["resource_name"] == resource_ref:
            fr = dict(fj, resource_path=os.path.join(fj["resource_dir"]+"/", resource_ref))
            return fr
        ft = make_turtle_resource_info(fj)
        if ft["resource_name"] == resource_ref:
            fr = dict(ft, resource_path=os.path.join(ft["resource_dir"]+"/", resource_ref))
            return fr
    log.debug("EntityResourceAccess.find_fixed_resource: %s not found"%(resource_ref))
    return None

def find_entity_resource(entity, resource_ref, fixed_resources=entity_fixed_json_resources):
    """
    Return a description for the indicated entity resource, or None

    resource_ref    is the local name of the desired resource relative to the 
                    entity to which it belongs.
    fixed_resources is a table of fixed resource information, not necessarily 
                    referenced by the entity itself.

    The description returned is a dictionary with the following keys:
        resource_name:      filename of resource (i.e. part of URI path after final "/")
        resource_dir:       directory of resource (i.e. part of URI path up to final "/")
        resource_path:      a file or URI path to the resource data relative to the
                            URI of the entity to which it belongs.
        resource_type:      content-type of resource
        resource_access:    optional: if present, specifies a function that returns an 
                            alternative representation of a JSON-LD data resource.
                            (e.g. see `turtle_resource_file`)
    """
    log.debug(
        "EntityResourceAccess.find_entity_resource %s/%s/%s"%
        (entity.get_type_id(), entity.get_id(), resource_ref)
        )
    fr = find_fixed_resource(fixed_resources, resource_ref)
    if fr:
        return fr
    # Look for resource description in entity data
    # @@TESTME
    for t, f in entity.enum_fields():
        log.debug("find_resource: t %s, f %r"%(t,f))
        if isinstance(f, dict):
            if f.get("resource_name", None) == resource_ref:
                f = dict(f, resource_path=resource_ref)
                return f
    return None

def find_list_resource(type_id, list_id, list_ref):
    """
    Return a description for the indicated entity resource, or None
    """
    log.debug(
        "EntityResourceAccess.find_list_resource %s/%s/%s"%
        (list_id, type_id, list_ref)
        )
    return find_fixed_resource(entity_list_json_resources, list_ref)

def get_resource_file(entity, resource_info, base_url):
    """
    Create a file object from which resource data can be read.

    resource_info   is a value returned by `find_fixed_resource` or `find_entity_resource`
    base_url        is a base URL that may be used to resolving relative references in the
                    JSON-LD data.

    Returns a pair of values: the file object, and a content-type string.
    """
    if "resource_access" in resource_info:
        # Use indicated resource access renderer
        jsondata = entity.get_values()
        resource_file = resource_info["resource_access"](base_url, jsondata, resource_info)
    else:
        # Return resource data direct from storage
        resource_file = entity.resource_file(resource_info["resource_path"])
    return (resource_file, resource_info["resource_type"])

# End.
